"""
Create a new instance of the `PyMarkdownApi` class.
"""

import argparse
import os
import tempfile
from dataclasses import dataclass
from typing import Any, List, Optional

from application_file_scanner import ApplicationFileScanner

from pymarkdown.application_logging import ApplicationLogging
from pymarkdown.general.main_presentation import MainPresentation
from pymarkdown.main import PyMarkdownLint
from pymarkdown.plugin_manager.plugin_scan_failure import PluginScanFailure

# pylint: disable=too-many-lines


# pylint: disable=too-many-instance-attributes,too-many-public-methods
class PyMarkdownApi:
    """
    Module to provide for an API to directly communicate with PyMarkdown instead
    of using a command line.

    Args:
        inherit_logging: If True, inherit the logging settings from the calling
            application.  If False, will use the `log_*` functions to specify the
            logging properties.
    """

    __INTERFACE_VERSION = 1

    def __init__(self, inherit_logging: bool = False) -> None:
        self.__inherit_logging = inherit_logging

        self.__log_level = ApplicationLogging.log_level_warning
        self.__log_file_path: Optional[str] = None
        self.__enable_stack_trace = False
        self.__enable_strict_configuration = False
        self.__plugin_paths_to_add: List[str] = []
        self.__configuration_path: Optional[str] = None
        self.__enable_rule_identifiers: List[str] = []
        self.__disable_rule_identifiers: List[str] = []
        self.__enable_extension_identifiers: List[str] = []
        self.__set_properties: List[str] = []
        self.__disable_json5_configuration = False
        self.__enable_continue_on_error = False

    # pylint: disable=too-many-arguments
    def scan_path(
        self,
        path_to_scan: str,
        recurse_if_directory: bool = False,
        alternate_extensions: Optional[str] = None,
        exclude_patterns: Optional[List[str]] = None,
        respect_gitignore: bool = False,
    ) -> "PyMarkdownScanPathResult":
        """
        Scan any eligible Markdown files found on the provided path.  For more information,
        check out our User's Guide sections on [Basic Scanning](../user-guide.md#basic-scanning)
        and [Command Line Arguments](../user-guide.md#command-line-arguments).

        Args:
            path_to_scan (str): If `path_to_scan` is a directory, scan within that directory for
                eligible Markdown files.  If `path_to_scan` is a file, determine if the file is
                an eligible Markdown file before proceeding with the scan.
            recurse_if_directory (bool): If `path_to_scan` is a directory and `recurse_if_directory`
                if ``True``, also scan any directories within the specified directory.
            alternate_extensions (str, optional): Optionally specify one or more comma-separated file
                extensions. Files with these file extensions are also considered to be eligible
                files to scan.
            exclude_patterns (List[str], optional): Optionally specify one or more glob-style patterns
                to exclude files or directories from being scanned.
            respect_gitignore (bool): If ``True``, respect any `.gitignore` files found when scanning
                according to standard Git rules.

        Raises:
            PyMarkdownApiArgumentException: If `path_to_scan` is empty or if `alternate_extensions`
                does not contain a valid list of file extensions.  Valid file extensions start with
                a single period character and are followed by one or more ASCII alphanumeric characters.
            PyMarkdownApiNoFilesFoundException: If no eligible files were found.
            PyMarkdownApiException: If some other error was found.

        Returns:
            An instance of `PyMarkdownScanPathResult` containing any scan failures or pragma errors
                encountered when scanning the eligible files on the provided path.

        Examples:
            This example scans a specific Markdown file `file.md` for any
            issues.  Both good results (printing of the two properties of `scan_result`) and
            bad results (printing the exception information) are properly handled.

                from pymarkdown.api import PyMarkdownApi, PyMarkdownApiException

                try:
                    scan_result = PyMarkdownApi().scan_path("file.md")
                    print(f"Scan Failures: {scan_result.scan_failures}")
                    print(f"Pragma Errors: {scan_result.pragma_errors}")
                except PyMarkdownApiException as this_exception:
                    print(f"API Exception: {this_exception}", file=sys.stderr)

            By replacing the line with `scan_path` with:

                scan_result = PyMarkdownApi().scan_path("./docs", recurse_if_directory=True)

            the code will now scan for any `*.md` files in the `./docs` directory and in
            any directories under the `./docs` directory.
        """
        self.__verify_string_argument_not_empty("path_to_scan", path_to_scan)

        scan_arguments = self.__build_common_arguments("scan")
        self.__add_common_scan_arguments(
            scan_arguments,
            path_to_scan,
            recurse_if_directory,
            alternate_extensions,
            exclude_patterns=exclude_patterns,
            respect_gitignore=respect_gitignore,
        )

        this_presentation = _ApiPresentation()
        scanner_instance = PyMarkdownLint(
            presentation=this_presentation,
            show_stack_trace=self.__enable_stack_trace,
            inherit_logging=self.__inherit_logging,
        )
        return_code = 0
        try:
            scanner_instance.main(scan_arguments)
        except SystemExit as this_exception:
            # https://github.com/python/typeshed/issues/8513#issue-1333671093
            return_code = (
                int(this_exception.code) if isinstance(this_exception.code, int) else 99
            )
            if return_code == 1:
                raise PyMarkdownApiNoFilesFoundException(
                    "No matching files found."
                ) from this_exception
        return self.__handle_scan_results(return_code, this_presentation)

    # pylint: enable=too-many-arguments

    def scan_string(
        self,
        string_to_scan: str,
    ) -> "PyMarkdownScanPathResult":
        """
        Scan the specified string as a Markdown document.

        Args:
            string_to_scan (str): String to interpret as a Markdown document.

        Raises:
            PyMarkdownApiArgumentException: If `string_to_scan` is empty.
            PyMarkdownApiException: If some other error was found.

        Returns:
            An instance of `PyMarkdownScanPathResult` containing any scan failures or pragma errors
                encountered when scanning the Markdown document.

        Examples:
            This example is a simplified form of the `scan_path` example that accepts a
            string to scan instead

                from pymarkdown.api import PyMarkdownApi, PyMarkdownApiException

                markdown_string = \"# Markdown\\n\\nIs cool!\\n\"

                try:
                    scan_result = PyMarkdownApi().scan_string(markdown_string)
                    print(f"Scan Failures: {scan_result.scan_failures}")
                    print(f"Pragma Errors: {scan_result.pragma_errors}")
                except PyMarkdownApiException as this_exception:
                    print(f"API Exception: {this_exception}", file=sys.stderr)
        """
        self.__verify_string_argument_not_empty("path_to_scan", string_to_scan)

        scan_arguments = self.__build_common_arguments("scan-stdin")

        this_presentation = _ApiPresentation()
        scanner_instance = PyMarkdownLint(
            presentation=this_presentation,
            show_stack_trace=self.__enable_stack_trace,
            inherit_logging=self.__inherit_logging,
            string_to_scan=string_to_scan,
        )
        return_code = 0
        try:
            scanner_instance.main(scan_arguments)
        except SystemExit as this_exception:
            # https://github.com/python/typeshed/issues/8513#issue-1333671093
            return_code = (
                int(this_exception.code) if isinstance(this_exception.code, int) else 99
            )
        return self.__handle_scan_results(return_code, this_presentation)

    # pylint: disable=too-many-arguments
    def fix_path(
        self,
        path_to_scan: str,
        recurse_if_directory: bool = False,
        alternate_extensions: Any = None,
        exclude_patterns: Optional[List[str]] = None,
        respect_gitignore: bool = False,
    ) -> "PyMarkdownFixResult":
        """
        Fix any eligible Markdown files found on the provided path that have scan failures that
        can be automatically fixed.

        Args:
            path_to_scan (str): If ``path_to_scan`` is a directory, scan within that directory for
                eligible Markdown files.  If `path_to_scan` is a file, determine if the file is
                an eligible Markdown file before proceeding with the scan and fix.
            recurse_if_directory (bool): If `path_to_scan` is a directory and `recurse_if_directory`
                if ``True``, also scan any directories within the specified directory.
            alternate_extensions (str, optional): Optionally specify one or more comma-separated file
                extensions. Files with these file extensions are also considered to be eligible
                files to scan.
            exclude_patterns (List[str], optional): Optionally specify one or more glob-style patterns
                to exclude files or directories from being scanned.
            respect_gitignore (bool): If ``True``, respect any `.gitignore` files found when scanning
                according to standard Git rules.

        Raises:
            PyMarkdownApiArgumentException: If `path_to_scan` is empty or if `alternate_extensions`
                does not contain a valid list of file extensions.  Valid file extensions start with
                a single period character and are followed by one or more ASCII alphanumeric characters.
            PyMarkdownApiNoFilesFoundException: If no eligible files were found.
            PyMarkdownApiException: If some other error was found.

        Returns:
            An instance containing any eligible files that were fixed.

        Examples:
            This function is syntactically equivalent to calling the `scan_path` function
            except for its return value.

                from pymarkdown.api import PyMarkdownApi, PyMarkdownApiException

                try:
                    fix_result = PyMarkdownApi().fix_path("file.md")
                    print(f"Fixed file paths: {fix_result.files_fixed}")
                except PyMarkdownApiException as this_exception:
                    print(f"API Exception: {this_exception}", file=sys.stderr)
        """
        self.__verify_string_argument_not_empty("path_to_scan", path_to_scan)

        scan_arguments = self.__build_common_arguments("fix")
        self.__add_common_scan_arguments(
            scan_arguments,
            path_to_scan,
            recurse_if_directory,
            alternate_extensions,
            exclude_patterns=exclude_patterns,
            respect_gitignore=respect_gitignore,
        )

        this_presentation = _ApiPresentation()
        scanner_instance = PyMarkdownLint(
            presentation=this_presentation,
            show_stack_trace=self.__enable_stack_trace,
            inherit_logging=self.__inherit_logging,
        )
        return_code = 0
        try:
            scanner_instance.main(scan_arguments)
        except SystemExit as this_exception:
            # https://github.com/python/typeshed/issues/8513#issue-1333671093
            return_code = (
                int(this_exception.code) if isinstance(this_exception.code, int) else 99
            )
        return self.__handle_fix_results(return_code, this_presentation)

    # pylint: enable=too-many-arguments

    def fix_string(
        self,
        string_to_scan: str,
    ) -> "PyMarkdownFixStringResult":
        """
        Scan the specified string as a Markdown document and apply any eligible fixes.

        Args:
            string_to_scan (str): String to interpret as a Markdown document.

        Raises:
            PyMarkdownApiArgumentException: If `string_to_scan` is empty.
            PyMarkdownApiException: If some other error was found.

        Returns:
            An instance containing the original document with any fixes applied.

        Examples:
            This function is syntactically equivalent to calling the `scan_string` function
            except for its return value.

                from pymarkdown.api import PyMarkdownApi, PyMarkdownApiException

                markdown_string = \"# Markdown\\n\\nIs cool!\\n\"

                try:
                    fix_result = PyMarkdownApi().fix_string(markdown_string)
                    print(f"Applied fixes?  {fix_result.was_fixed}")
                    print(f"Fixed Markdown: {fix_result.fixed_file}")
                except PyMarkdownApiException as this_exception:
                    print(f"API Exception: {this_exception}", file=sys.stderr)
        """
        self.__verify_string_argument_not_empty("string_to_scan", string_to_scan)

        temp_file = None
        try:
            with tempfile.NamedTemporaryFile(
                "wt", suffix=".md", encoding="utf-8", delete=False
            ) as temp_file:
                temp_file.write(string_to_scan)

            scan_arguments = self.__build_common_arguments("fix")
            scan_arguments.append(temp_file.name)

            this_presentation = _ApiPresentation()
            scanner_instance = PyMarkdownLint(
                presentation=this_presentation,
                show_stack_trace=self.__enable_stack_trace,
                inherit_logging=self.__inherit_logging,
            )
            return_code = 0
            try:
                scanner_instance.main(scan_arguments)
            except SystemExit as this_exception:
                # https://github.com/python/typeshed/issues/8513#issue-1333671093
                return_code = (
                    int(this_exception.code)
                    if isinstance(this_exception.code, int)
                    else 99
                )
            fix_result = self.__handle_fix_results(return_code, this_presentation)
            with open(temp_file.name, "rt", encoding="utf-8") as fixed_file:
                return PyMarkdownFixStringResult(
                    bool(fix_result.files_fixed), fixed_file.read()
                )
        finally:
            if temp_file and os.path.isfile(temp_file.name):  # pragma: no cover
                os.remove(temp_file.name)

    # pylint: disable=too-many-arguments
    def list_path(
        self,
        path_to_scan: str,
        recurse_if_directory: bool = False,
        alternate_extensions: str = "",
        exclude_patterns: Optional[List[str]] = None,
        respect_gitignore: bool = False,
    ) -> "PyMarkdownListPathResult":
        """
        List any eligible files found when scanning the specified path for eligible markdown files.
        This function is provided for debugging situations to provide confidence that one or more
        Markdown files are indeed being scanned by PyMarkdown.

        Args:
            path_to_scan (str): If ``path_to_scan`` is a directory, scan within that directory for
                eligible Markdown files.  If `path_to_scan` is a file, determine if the file is
                an eligible Markdown file.
            recurse_if_directory (bool): If `path_to_scan` is a directory and `recurse_if_directory`
                if ``True``, also scan any directories within the specified directory.
            alternate_extensions (str, optional): Optionally specify one or more comma-separated file
                extensions. Files with these file extensions are also considered to be eligible
                files to scan.
            exclude_patterns (List[str], optional): Optionally specify one or more glob-style patterns
                to exclude files or directories from being scanned.
            respect_gitignore (bool): If ``True``, respect any `.gitignore` files found when scanning
                according to standard Git rules.

        Raises:
            PyMarkdownApiArgumentException: If `path_to_scan` is empty or if `alternate_extensions`
                does not contain a valid list of file extensions.  Valid file extensions start with
                a single period character and are followed by one or more ASCII alphanumeric characters.
            PyMarkdownApiNoFilesFoundException: If no eligible files were found.
            PyMarkdownApiException: If some other error was found.

        Returns:
            An instance containing eligible Markdown files that would have normally be scanned.

        Examples:
            This example generates a list of the files that PyMarkdown will scan if the `scan_path`
            function is invoked with the same parameters.

                from pymarkdown.api import PyMarkdownApi, PyMarkdownApiException

                try:
                    scan_result = PyMarkdownApi().list_path("./docs", recurse_if_directory=True)
                    print(scan_result.matching_files)
                except PyMarkdownApiException as this_exception:
                    print(f"API Exception: {this_exception}", file=sys.stderr)
        """
        self.__verify_string_argument_not_empty("path_to_scan", path_to_scan)

        scan_arguments = self.__build_common_arguments("scan")
        scan_arguments.append("--list-files")
        self.__add_common_scan_arguments(
            scan_arguments,
            path_to_scan,
            recurse_if_directory,
            alternate_extensions,
            exclude_patterns=exclude_patterns,
            respect_gitignore=respect_gitignore,
        )

        this_presentation = _ApiPresentation()
        scanner_instance = PyMarkdownLint(
            presentation=this_presentation,
            show_stack_trace=self.__enable_stack_trace,
            inherit_logging=self.__inherit_logging,
        )
        return_code = 0
        try:
            scanner_instance.main(scan_arguments)
        except SystemExit as this_exception:
            return_code = (
                int(this_exception.code) if isinstance(this_exception.code, int) else 99
            )

        if return_code != 0:
            self.__generate_scan_exception(this_presentation)
        return PyMarkdownListPathResult(this_presentation.pso[0].split("\n"))

    # pylint: enable=too-many-arguments

    @property
    def application_version(self) -> str:
        """
        Report on the application version.

        Returns:
            The current application version.

        Examples:
            This function queries the current version of PyMarkdown.

                from pymarkdown.api import PyMarkdownApi

                print(f"PyMarkdown version = {PyMarkdownApi().application_version}")
        """
        return PyMarkdownLint().application_version

    @property
    def interface_version(self) -> int:
        """
        Report on the interface version.

        Returns:
            The current plugin interface version.

        Examples:
            This function queries the current version of this API.

                from pymarkdown.api import PyMarkdownApi

                print(f"PyMarkdown API version = {PyMarkdownApi().interface_version}")
        """
        return PyMarkdownApi.__INTERFACE_VERSION

    def disable_rule_by_identifier(self, rule_identifier: str) -> "PyMarkdownApi":
        """
        Disable a single rule by one of its identifiers.

        Args:
            rule_identifier (str): Identifier for the rule to disable.

        Raises:
            PyMarkdownApiArgumentException: If `rule_identifier` is empty.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function disables a single rule by one of its identifiers.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().disable_rule_by_identifier("md031").scan_path("file.md")
        """
        self.__verify_string_argument_not_empty("rule_identifier", rule_identifier)

        self.__disable_rule_identifiers.append(rule_identifier)
        return self

    def enable_rule_by_identifier(self, rule_identifier: str) -> "PyMarkdownApi":
        """
        Enable a single rule by one of its identifiers.

        Args:
            rule_identifier (str): Identifier for the rule to enable.

        Raises:
            PyMarkdownApiArgumentException: If `rule_identifier` is empty.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function enables a single rule by one of its identifiers.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().enable_rule_by_identifier("md031").scan_path("file.md")
        """
        self.__verify_string_argument_not_empty("rule_identifier", rule_identifier)

        self.__enable_rule_identifiers.append(rule_identifier)
        return self

    def enable_extension_by_identifier(
        self, extension_identifier: str
    ) -> "PyMarkdownApi":
        """
        Enable a single extension by its identifier.

        Args:
            extension_identifier (str): Identifier for the extension to enable.

        Raises:
            PyMarkdownApiArgumentException: If `extension_identifier` is empty.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function enables a single extension by its identifier.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().enable_extension_by_identifier("front-matter").scan_path("file.md")
        """
        self.__verify_string_argument_not_empty(
            "extension_identifier", extension_identifier
        )

        self.__enable_extension_identifiers.append(extension_identifier)
        return self

    def configuration_file_path(self, path_to_config_file: str) -> "PyMarkdownApi":
        """
        Set the path to the configuration file to use.

        Args:
            path_to_config_file (str): Path to the configuration file to use.

        Raises:
            PyMarkdownApiArgumentException: If `path_to_config_file` is empty.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function specifies a file containing configuration information.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().path_to_config_file("pymarkdown.cfg").scan_path("file.md")
        """
        self.__verify_string_argument_not_empty(
            "path_to_config_file", path_to_config_file
        )

        self.__configuration_path = path_to_config_file
        return self

    def set_boolean_property(
        self, property_name: str, property_value: bool
    ) -> "PyMarkdownApi":
        """
        Set a named configuration property to a given boolean value.

        Args:
            property_name (str): Full hierarchical name of the property to set.
            property_value (bool): `True` or `False` boolean value to set the property to.

        Raises:
            PyMarkdownApiArgumentException: If `property_name` is empty or if `property_value` is not a `bool` value.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function sets a boolean value for a specific configuration item.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().set_boolean_property("plugins.md007.enabled", True).scan_path("file.md")
        """
        self.__verify_string_argument_not_empty("property_name", property_name)
        if not isinstance(property_value, bool):
            raise PyMarkdownApiArgumentException(
                "property_value",
                "The property value 'property_value' was not passed as a boolean.",
            )

        self.__set_properties.append(f"{property_name}=$!{property_value}")
        return self

    def set_integer_property(
        self, property_name: str, property_value: int
    ) -> "PyMarkdownApi":
        """
        Set a named configuration property to a given integer value.

        Args:
            property_name (str): Full hierarchical name of the property to set.
            property_value (int): Integer value to set the property to.

        Raises:
            PyMarkdownApiArgumentException: If `property_name` is empty or if `property_value` is not an `int` value.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function sets a integer value for a specific configuration item.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().set_integer_property("plugins.md007.code_block_line_length", 160).scan_path("file.md")
        """
        self.__verify_string_argument_not_empty("property_name", property_name)
        if not isinstance(property_value, int):
            raise PyMarkdownApiArgumentException(
                "property_value",
                "The property value 'property_value' was not passed as an integer.",
            )

        self.__set_properties.append(f"{property_name}=$#{property_value}")
        return self

    def set_string_property(
        self, property_name: str, property_value: str
    ) -> "PyMarkdownApi":
        """
        Set a named configuration property to a given string value.

        Args:
            property_name (str): Full hierarchical name of the property to set.
            property_value (str): String value to set the property to.

        Raises:
            PyMarkdownApiArgumentException: If `property_name` is empty or if `property_value` is not an `str` value.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function sets a string value for a specific configuration item.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().set_string_property("plugins.heading-style-h1.style", "consistent").scan_path("file.md")
        """
        self.__verify_string_argument_not_empty("property_name", property_name)
        if not isinstance(property_value, str):
            raise PyMarkdownApiArgumentException(
                "property_value",
                "The property value 'property_value' was not passed as a string.",
            )

        self.__set_properties.append(f"{property_name}=$${property_value}")
        return self

    def set_property(self, property_name: str, property_value: Any) -> "PyMarkdownApi":
        """
        Set a named configuration property to a given value.  Whatever is passed in as
        the property_value parameter is translated into a string.

        If at all possible, the other three `set_x_property` functions should be used
        as they ensure that the correct type of value is set as the configuration item
        value.  This function translated ANY value for the `property_value` argument
        into a `str` before setting it as the configuration item value.

        Args:
            property_name (str): Full hierarchical name of the property to set.
            property_value (Any): Value to set the property to after applying a string
                transformation to the value.

        Raises:
            PyMarkdownApiArgumentException: If `property_name` is empty.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function sets a value, translated into a string, for a specific configuration item.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().set_property("plugins.heading-style-h1.style", 1).scan_path("file.md")
        """
        self.__verify_string_argument_not_empty("property_name", property_name)

        self.__set_properties.append(f"{property_name}={str(property_value)}")
        return self

    def enable_strict_configuration(self) -> "PyMarkdownApi":
        """
        Enable strict configuration for any requested properties, either through configuration files or manual setting.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function enforces strict adherence to configuration requirements.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().enable_strict_configuration().scan_path("file.md")
        """
        self.__enable_strict_configuration = True
        return self

    def log_debug_and_above(self) -> "PyMarkdownApi":
        """
        Enable logging for the DEBUG level and above.

        Raises:
            PyMarkdownApiNotSupportedException: If invoked after `inherit_logging` was set
                when creating the `PyMarkdownApi` instance.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function sets the logging level for PyMarkdown to anything that
            is at `DEBUG` logging level or above.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().log_debug_and_above().scan_path("file.md")
        """
        return self.log(ApplicationLogging.log_level_debug)

    def log_info_and_above(self) -> "PyMarkdownApi":
        """
        Enable logging for the INFO level and above.

        Raises:
            PyMarkdownApiNotSupportedException: If invoked after `inherit_logging` was set
                when creating the `PyMarkdownApi` instance.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function sets the logging level for PyMarkdown to anything that
            is at `INFO` logging level or above.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().log_info_and_above().scan_path("file.md")
        """
        return self.log(ApplicationLogging.log_level_info)

    def log_warning_and_above(self) -> "PyMarkdownApi":
        """
        Enable logging for the WARN level and above.

        Raises:
            PyMarkdownApiNotSupportedException: If invoked after `inherit_logging` was set
                when creating the `PyMarkdownApi` instance.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function sets the logging level for PyMarkdown to anything that
            is at `WARNING` logging level or above.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().log_warning_and_above().scan_path("file.md")
        """
        return self.log(ApplicationLogging.log_level_warning)

    def log_error_and_above(self) -> "PyMarkdownApi":
        """
        Enable logging for the ERROR level and above.

        Raises:
            PyMarkdownApiNotSupportedException: If invoked after `inherit_logging` was set
                when creating the `PyMarkdownApi` instance.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function sets the logging level for PyMarkdown to anything that
            is at `ERROR` logging level or above.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().log_error_and_above().scan_path("file.md")
        """
        return self.log(ApplicationLogging.log_level_error)

    def log_critical_and_above(self) -> "PyMarkdownApi":
        """
        Enable logging for the CRITICAL level and above.

        Raises:
            PyMarkdownApiNotSupportedException: If invoked after `inherit_logging` was set
                when creating the `PyMarkdownApi` instance.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function sets the logging level for PyMarkdown to anything that
            is at `CRITICAL` logging level or above.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().log_critical_and_above().scan_path("file.md")
        """
        return self.log(ApplicationLogging.log_level_critical)

    def log(self, log_level: str) -> "PyMarkdownApi":
        """
        Set the logging level using a string value.

        Args:
            log_level (str): One of "CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG".

        Raises:
            PyMarkdownApiArgumentException: If `log_level` is not one of the allowed values.
            PyMarkdownApiNotSupportedException: If invoked after `inherit_logging` was set
                when creating the `PyMarkdownApi` instance.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function sets the logging level for PyMarkdown to anything that
            is at `INFO` logging level or above.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().log("INFO").scan_path("file.md")
        """
        if not ApplicationLogging.is_valid_log_level_type(log_level):
            log_levels_in_order = ",".join(ApplicationLogging.get_valid_log_levels())
            raise PyMarkdownApiArgumentException(
                "log_level",
                f"Parameter 'log_level' must be one of {log_levels_in_order}",
            )

        if self.__inherit_logging:
            raise PyMarkdownApiNotSupportedException(
                "Set log level functions are not supported in log-inheritance mode."
            )

        self.__log_level = log_level
        return self

    def log_to_file(self, log_file_path: str) -> "PyMarkdownApi":
        """
        Set a file to log any results to.

        Args:
            log_file_path (str): Path to the file to write the logs to.

        Raises:
            PyMarkdownApiArgumentException: If `log_file_path` is empty.
            PyMarkdownApiNotSupportedException: If invoked after `inherit_logging` was set
                when creating the `PyMarkdownApi` instance.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function writes any logging messages to the specified file.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().log_to_file("file.log").scan_path("file.md")
        """
        self.__verify_string_argument_not_empty("log_file_path", log_file_path)

        if self.__inherit_logging:
            raise PyMarkdownApiNotSupportedException(
                "Set log file function is not supported in log-inheritance mode."
            )

        self.__log_file_path = log_file_path
        return self

    def add_plugin_path(self, path_to_plugin: str) -> "PyMarkdownApi":
        """
        Add a plugin path that points to a directory with plugins or a single plugin.

        Args:
            path_to_plugin (str): Path to an additional plugin to load.

        Raises:
            PyMarkdownApiArgumentException: If `path_to_plugin` is empty.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function informs PyMarkdown to use a new plugin.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().add_plugin_path("my_plugin.py").scan_path("file.md")
        """
        self.__verify_string_argument_not_empty("path_to_plugin", path_to_plugin)

        self.__plugin_paths_to_add.append(path_to_plugin)
        return self

    def enable_stack_trace(self) -> "PyMarkdownApi":
        """
        Enable the reporting of stack traces for any exceptions caught by the API.
        If this modifier is present, additional information will be recorded as
        to the exact nature of any reported exception.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function enable stack trace support in case of a raised exception.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().enable_stack_trace().scan_path("file.md")
        """
        self.__enable_stack_trace = True
        return self

    def disable_json5_configuration(self) -> "PyMarkdownApi":
        """
        Disable the use of the JSON5 parser for configuration files, instead
        using the base json parser from the Python standard library.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function disables the use of the JSON5 parser for configuration files.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().disable_json5_configuration().scan_path("file.md")
        """
        self.__disable_json5_configuration = True
        return self

    def enable_continue_on_error(self) -> "PyMarkdownApi":
        """
        Enable the scanning of multiple files to continue, even if some of the files
        have critical errors.

        Returns:
            An instance of `PyMarkdownApi` to allow for function chaining.

        Examples:
            This function disables the use of the JSON5 parser for configuration files.

                from pymarkdown.api import PyMarkdownApi

                PyMarkdownApi().disable_json5_configuration().scan_path("file.md")
        """
        self.__enable_continue_on_error = True
        return self

    def __handle_scan_results(
        self, return_code: int, this_presentation: "_ApiPresentation"
    ) -> "PyMarkdownScanPathResult":
        assert (
            not this_presentation.pso
        ), "should not display for scan_path, but for ext ops and plugin ops"
        if return_code != 0 and not self.__enable_continue_on_error:
            self.__generate_scan_exception(this_presentation)
        return PyMarkdownScanPathResult(
            this_presentation.scan_failures,
            this_presentation.pragma_errors,
            this_presentation.pse,
        )

    def __handle_fix_results(
        self, return_code: int, this_presentation: "_ApiPresentation"
    ) -> "PyMarkdownFixResult":
        assert (
            not this_presentation.pso
        ), "should not display for scan_path, but for ext ops and plugin ops"
        if return_code not in [0, 1, 3] and not self.__enable_continue_on_error:
            raise PyMarkdownApiException(this_presentation.pse[-1].strip("\n"))
        return PyMarkdownFixResult(this_presentation.files_fixed, this_presentation.pse)

    def __generate_scan_exception(self, this_presentation: "_ApiPresentation") -> None:
        if not this_presentation.pse:
            return
        if this_presentation.pse[-1] == "\n\nNo matching files found.":
            raise PyMarkdownApiNoFilesFoundException("No matching files found.")
        raise PyMarkdownApiException(this_presentation.pse[-1].strip("\n"))

    # pylint: disable=too-many-arguments
    def __add_common_scan_arguments(
        self,
        scan_arguments: List[str],
        path_to_scan: str,
        recurse_if_directory: bool,
        alternate_extensions: Optional[str],
        exclude_patterns: Optional[List[str]] = None,
        respect_gitignore: bool = False,
    ) -> None:
        if recurse_if_directory:
            scan_arguments.append("--recurse")

        if respect_gitignore:
            scan_arguments.append("--respect-gitignore")

        if alternate_extensions:
            self.__verify_string_argument_alternate_extensions(
                "alternate_extensions", alternate_extensions
            )
            scan_arguments.extend(("-ae", alternate_extensions))

        if exclude_patterns:
            for next_pattern in exclude_patterns:
                scan_arguments.extend(("--exclude", next_pattern))

        scan_arguments.append(path_to_scan)

    # pylint: enable=too-many-arguments

    def __build_common_arguments(self, action_to_invoke: str) -> List[str]:

        # Note: `--return-code-scheme` is not included as a configurable option
        #       as the api requires explicit information about what happened.
        common_arguments: List[str] = ["--return-code-scheme", "explicit"]

        if self.__disable_json5_configuration:
            common_arguments.append("--no-json5")
        if self.__enable_continue_on_error:
            common_arguments.append("--continue-on-error")
        if self.__enable_stack_trace:
            common_arguments.append("--stack-trace")
        if self.__enable_strict_configuration:
            common_arguments.append("--strict-config")

        if not self.__inherit_logging:
            if self.__log_file_path:
                common_arguments.extend(("--log-file", self.__log_file_path))
            common_arguments.extend(["--log-level", self.__log_level])

        if self.__configuration_path:
            common_arguments.extend(["--config", self.__configuration_path])

        for next_property in self.__set_properties:
            common_arguments.extend(("--set", next_property))
        for next_path in self.__plugin_paths_to_add:
            common_arguments.extend(("--add-plugin", next_path))
        if self.__enable_rule_identifiers:
            rules_to_enable = "".join(
                f",{next_identifier}"
                for next_identifier in self.__enable_rule_identifiers
            )
            common_arguments.extend(("--enable-rules", rules_to_enable))
        if self.__disable_rule_identifiers:
            rules_to_disable = "".join(
                f",{next_identifier}"
                for next_identifier in self.__disable_rule_identifiers
            )
            common_arguments.extend(("--disable-rules", rules_to_disable))
        if self.__enable_extension_identifiers:
            common_arguments.extend(
                ("--enable-extensions", ",".join(self.__enable_extension_identifiers))
            )

        common_arguments.append(action_to_invoke)
        return common_arguments

    def __verify_string_argument_not_empty(
        self, argument_name: str, string_to_validate: str
    ) -> None:
        string_to_validate = string_to_validate.strip()
        if not string_to_validate:
            raise PyMarkdownApiArgumentException(
                argument_name, f"Parameter named '{argument_name}' cannot be empty."
            )

    def __verify_string_argument_alternate_extensions(
        self, argument_name: str, string_to_validate: str
    ) -> None:
        try:
            _ = ApplicationFileScanner.is_valid_comma_separated_extension_list(
                string_to_validate
            )
        except argparse.ArgumentTypeError as this_exception:
            raise PyMarkdownApiArgumentException(
                argument_name,
                f"Parameter named '{argument_name}' is not a valid comma-separated list of extensions.",
            ) from this_exception


# pylint: enable=too-many-instance-attributes,too-many-public-methods


@dataclass(frozen=True)
class PyMarkdownScanFailure:
    """
    Class to contain information about a failure reported by one of the rule plugins.

    Each instance of this class specifies a single scan failure that was encountered
    within the specified Markdown document.  Using the encapsulated fields, the following
    pieces of information about that scan failure can be determined:

    - the location where the scan failure occurred
    - the specific location within that location that triggered the scan failure
    - the exact rule that was triggered

    Optionally, depending on the rule, extra information may be present in the
    `extra_error_information` field.  This information is rule-dependent and is
    intended to provide extra clarity on why the rule was triggered.

    For more information consult the [Rule Failure Format](../user-guide.md#rule-failure-format)
    section of our User's Guide.

    Attributes:
        scan_file (str): Path to the file containing the failure.
        line_number (int): Line number of the triggered rule failure.
        column_number (int): Column number of the triggered rule failure.
        rule_id (str): ID of the rule that was triggered.
        rule_name (str): Name(s) of the rule that was triggered.
        rule_description (str): Longer description of the rule that was triggered.
        extra_error_information (Optional[str]): String providing more information on why the rule was triggered.
    """

    scan_file: str
    """File that was being scanned when the failure occurred."""
    line_number: int
    """Line number of the triggered rule failure."""
    column_number: int
    """Column number of the triggered rule failure."""
    rule_id: str
    """ID of the rule that was triggered."""
    rule_name: str
    """Name(s) of the rule that was triggered."""
    rule_description: str
    """Longer description of the rule that was triggered."""
    extra_error_information: Optional[str]
    """Optional string providing more information on why the rule was triggered."""

    def partial_equals(self, other: "PyMarkdownScanFailure") -> bool:
        """
        Decide if special fields are the same from both items.
        """
        return (
            self.scan_file == other.scan_file
            and self.line_number == other.line_number
            and self.column_number == other.column_number
            and self.rule_id == other.rule_id
        )


@dataclass(frozen=True)
class PyMarkdownPragmaError:
    """
    Class to encapsulate the information for a pragma error.

    Each instance of this class specifies a single pragma failure that was encountered
    within the specified Markdown document.  Using the encapsulated fields, the following
    pieces of information about that pragma error can be determined:

    - the location where the pragma error occurred
    - the specific location within that location that triggered the pragma error
    - specific information on why PyMarkdown generated the pragma error

    More information on Pragmas and
    their use are available [here](../user-guide.md#pragma-extension).

    Attributes:
        file_path (str): Path to the file that contains the improperly constructed pragma.
        line_number (int): Line number where the pragma is contained.
        pragma_error (str): Specific information about the error.
    """

    file_path: str
    """Path to the file that contains the improperly constructed pragma."""
    line_number: int
    """Line number where the pragma is contained."""
    pragma_error: str
    """Specific information about the error."""


@dataclass(frozen=True)
class PyMarkdownScanPathResult:
    """
    Result for the `scan_path` and `scan_string` functions.

    As both `PyMarkdownScanFailure` objects and `PyMarkdownPragmaError` objects
    contain location and additional failure/error information, this result
    object is a simple pair of lists containing failure information.

    Attributes:
        scan_failures (List[PyMarkdownScanFailure]): Zero or more `PyMarkdownScanFailure` objects.
        pragma_errors (List[PyMarkdownPragmaError]): Zero or more `PyMarkdownPragmaError` objects.
    """

    scan_failures: List[PyMarkdownScanFailure]
    """
    List of zero or more `PyMarkdownScanFailure` objects.
    """
    pragma_errors: List[PyMarkdownPragmaError]
    """
    List of zero or more `PyMarkdownPragmaError` objects.
    """
    critical_errors: List[str]
    """
    List of zero or more critical errors that were encountered during the scan. Only
    set if `enable_continue_on_error` was set when the `scan_path` or `scan_string`
    function was invoked.  If no critical errors were encountered, this list is empty.
    """


@dataclass(frozen=True)
class PyMarkdownFixResult:
    """
    Result for the fix_path function.

    The only information that PyMarkdown provides about scanned and fixed
    documents are the names of the documents that were fixed.  As such, this result
    simply provides those same Markdown file names.

    Attributes:
        files_fixed (List[str]): List of zero or more files that were fixed.
    """

    files_fixed: List[str]
    """
    List of zero or more files that were fixed.
    """
    critical_errors: List[str]
    """
    List of zero or more critical errors that were encountered during the fixing of the files. Only
    set if `enable_continue_on_error` was set when the `fix_path` function was invoked.
    If no critical errors were encountered, this list is empty.
    """


@dataclass(frozen=True)
class PyMarkdownFixStringResult:
    """
    Result for the fix_string function.

    Focusing on a singular Markdown document, this object returns an indication
    of whether fixes were applied along with the fixed document.

    Attributes:
        was_fixed (bool): Whether the string, interpretted as a Markdown document, was fixed.
        fixed_file (str): String that was passed into the `fix_string` function, with any fixes applied to it.
    """

    was_fixed: bool
    """
    Whether the string was fixed.
    """

    fixed_file: str
    """
    String that was passed into the `fix_string` function, with any fixes applied
    to it.
    """


@dataclass(frozen=True)
class PyMarkdownListPathResult:
    """
    Result for the list_path function.

    This object returns a list of files that PyMarkdown understands to be eligible
    for scanning, without having scanned those files.

    Attributes:
        matching_files (List[str]): List of filenames that match the specifications of the requested path.
    """

    matching_files: List[str]
    """List of filenames that match the specifications of the requested path."""


class PyMarkdownApiException(Exception):
    """
    Class to provide for an exception that is thrown by the API layer.

    This base PyMarkdown application exception is explicitly thrown when a unexpected error occurs that
    is not categorized as one of the other application exceptions.  Where possible, child classes of
    this class should be used to provide more specific information regarding the nature of the
    exception.

    Attributes:
        reason (str): Reported reason why the action failed.
    """

    def __init__(self, reason: str) -> None:
        self.__reason = reason

    @property
    def reason(self) -> str:
        """
        Reported reason why the action failed.
        """
        return self.__reason


class PyMarkdownApiNotSupportedException(PyMarkdownApiException):
    """
    Class to provide for an exception that a given situation is not supported.

    The most frequent raising of this exception is when the `PyMarkdownApi` instance
    is created with the *inherit_logging* set to `True`, followed by a call to
    one of the functions that alters the logging behavior of the application.

    Attributes:
        reason (str): Reported reason why the action failed.
    """


class PyMarkdownApiArgumentException(PyMarkdownApiException):
    """
    Class to provide for an argument that an exception is not valid.

    This exception is raised when an argument to a function falls outside of the
    expected behavior for that function.  This can mean that a `str` parameter
    is unexpectedly empty or that the parameter is not one of the allowed values
    for that parameter.

    Attributes:
        reason (str): Reported reason why the action failed.
        argument_name(str): Name of the argument that caused this exception to be raised.
    """

    def __init__(self, argument_name: str, reason: str) -> None:
        super().__init__(reason)
        self.__argument_name = argument_name

    @property
    def argument_name(self) -> str:
        """
        Name of the argument that caused this exception to be raised.
        """
        return self.__argument_name


class PyMarkdownApiNoFilesFoundException(PyMarkdownApiException):
    """
    Class to provide for an exception that the invoked API was not able to find at least one file to process.

    This is raised as an exception if at least one file to scan or fix is
    not encountered.  As the PyMarkdown application is a scanning application,
    not finding a single Markdown document to scan is typically an exceptional
    case.

    Attributes:
        reason (str): Reported reason why the action failed.
    """


class _ApiPresentation(MainPresentation):
    """
    Class to provide for the output of the PyMarkdown application.
    """

    def __init__(self) -> None:
        self.pso: List[str] = []
        self.pse: List[str] = []
        self.pragma_errors: List[PyMarkdownPragmaError] = []
        self.scan_failures: List[PyMarkdownScanFailure] = []
        self.files_fixed: List[str] = []

    def print_system_output(self, output_string: str) -> None:
        """
        Root function to output to standard out.
        """
        self.pso.append(output_string)

    def print_system_error(self, error_string: str) -> None:
        """
        Root function to output to standard error.
        """
        self.pse.append(error_string)

    # def format_scan_error(
    #     self,
    #     next_file: str,
    #     this_exception: Exception,
    #     show_extended_information: bool = False,
    # ) -> Optional[str]:
    #     """
    #     Format a scan error for display.  Returning a value of None means that
    #     the function has handled any required output.
    #     """
    #     return super().format_scan_error(
    #         next_file, this_exception, show_extended_information
    #     )

    def print_pragma_failure(
        self, scan_file: str, line_number: int, pragma_error: str
    ) -> None:
        """
        Print a failure to compile the pragma.
        """
        self.pragma_errors.append(
            PyMarkdownPragmaError(scan_file, line_number, pragma_error)
        )

    def print_scan_failure(self, scan_failure: PluginScanFailure) -> None:
        """
        Print a scan failure for a specific file and location.
        """
        local_copy = PyMarkdownScanFailure(
            scan_file=scan_failure.scan_file,
            line_number=scan_failure.line_number,
            column_number=scan_failure.column_number,
            rule_id=scan_failure.rule_id,
            rule_name=scan_failure.rule_name,
            rule_description=scan_failure.rule_description,
            extra_error_information=scan_failure.extra_error_information,
        )
        self.scan_failures.append(local_copy)

    def print_fix_message(self, file_fixed: str) -> None:
        """
        Print a message indicating that a given file has been fixed.
        """
        self.files_fixed.append(file_fixed)



"""
Module to provide for a leaf element that can be added to markdown parsing stream that handles front matter.
"""

import logging
from typing import Dict, List, Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)

POGGER = ParserLogger(logging.getLogger(__name__))


class FrontMatterMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the front matter data.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        start_boundary_line: str,
        end_boundary_line: str,
        collected_lines: List[str],
        matter_map: Dict[str, str],
        position_marker: PositionMarker,
    ) -> None:
        self.__start_boundary_line = start_boundary_line
        self.__end_boundary_line = end_boundary_line
        self.__collected_lines = collected_lines
        self.__matter_map = matter_map

        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_front_matter,
            "",
            position_marker=position_marker,
            is_extension=True,
        )
        self.__compose_extra_data_field()

    # pylint: enable=too-many-arguments
    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_front_matter

    # pylint: enable=protected-access

    @property
    def start_boundary_line(self) -> str:
        """
        Returns the boundary line used to start the front matter block.
        """
        return self.__start_boundary_line

    @property
    def end_boundary_line(self) -> str:
        """
        Returns the boundary line used to stop the front matter block.
        """
        return self.__end_boundary_line

    @property
    def collected_lines(self) -> List[str]:
        """
        Returns the collected lines that comprise the front matter block.
        """
        return self.__collected_lines

    @property
    def matter_map(self) -> Dict[str, str]:
        """
        Returns the processed lines from the front matter block.
        """
        return self.__matter_map

    def __compose_extra_data_field(self) -> None:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        self._set_extra_data(
            MarkdownToken.extra_data_separator.join(
                [
                    self.__start_boundary_line,
                    self.__end_boundary_line,
                    str(self.__collected_lines),
                    str(self.__matter_map),
                ]
            )
        )

    @classmethod
    def calculate_block_token_height(cls, last_token: MarkdownToken) -> int:
        """
        Calculate the height of the token with the given properties.
        """
        front_matter_token = cast(FrontMatterMarkdownToken, last_token)
        return 2 + len(front_matter_token.collected_lines)

    @classmethod
    def calculate_initial_whitespace(cls) -> Tuple[int, bool]:
        """
        Calculate the amount of whitespace for the token.
        """
        return 0, False

    def register_for_markdown_transform(
        self,
        registration_function: RegisterMarkdownTransformHandlersProtocol,
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            FrontMatterMarkdownToken,
            FrontMatterMarkdownToken.__rehydrate_front_matter,
            None,
        )

    @staticmethod
    def __rehydrate_front_matter(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the front matter text from the token.
        """
        _ = previous_token, context

        front_mater_token = cast(FrontMatterMarkdownToken, current_token)
        front_matter_parts = [front_mater_token.start_boundary_line]
        front_matter_parts.extend(front_mater_token.collected_lines)
        front_matter_parts.extend([front_mater_token.end_boundary_line, ""])
        return ParserHelper.newline_character.join(front_matter_parts)

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            FrontMatterMarkdownToken,
            FrontMatterMarkdownToken.__handle_front_matter_token,
            None,
        )

    @staticmethod
    def __handle_front_matter_token(
        output_html: str, next_token: MarkdownToken, transform_state: TransformState
    ) -> str:
        """
        Handle the front matter token.  Note that it does not contribute anything
        at all to the HTML output.
        """
        _ = (next_token, transform_state)

        return output_html



"""
Module to provide processing for the nested scenarios that may contain container blocks.
"""

from __future__ import annotations

import logging
from typing import Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_count_helper import BlockQuoteCountHelper
from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.container_blocks.container_grab_bag import ContainerGrabBag
from pymarkdown.container_blocks.container_indices import ContainerIndices
from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.list_blocks.list_block_starts_helper import ListBlockStartsHelper
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.setext_heading_markdown_token import SetextHeadingMarkdownToken
from pymarkdown.tokens.stack_token import ListStackToken

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods
class ContainerBlockNestedProcessor:
    """
    Class to provide processing for the nested scenarios that may contain container blocks.
    """

    @staticmethod
    def handle_nested_container_blocks(
        parser_state: ParserState,
        position_marker: PositionMarker,
        was_container_start: bool,
        avoid_block_starts: bool,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[Optional[str], bool, bool]:
        """
        Handle the processing of nested container blocks, as they can contain
        themselves and get somewhat messy.
        """
        adjusted_text_to_parse = position_marker.text_to_parse

        POGGER.debug("adjusted_text_to_parse>$<", adjusted_text_to_parse)
        POGGER.debug("index_number>$<", position_marker.index_number)
        POGGER.debug("index_indent>$<", position_marker.index_indent)
        POGGER.debug("parser_state.nested_list_start>$", parser_state.nested_list_start)
        POGGER.debug("was_container_start>$", was_container_start)

        if was_container_start and position_marker.text_to_parse:
            assert (
                grab_bag.container_depth < 10
            ), "Prevent nesting from going beyond 10 deep."
            nested_container_starts = (
                ContainerBlockNestedProcessor.__get_nested_container_starts(
                    parser_state,
                    position_marker.text_to_parse,
                    avoid_block_starts,
                )
            )
            POGGER.debug(
                "__handle_nested_container_blocks>nested_container_starts>>:$:<<",
                nested_container_starts,
            )

            if (
                len(grab_bag.container_tokens) == 1
                and grab_bag.container_tokens[0].is_block_quote_end
            ):
                parser_state.token_document.extend(grab_bag.container_tokens)
                grab_bag.container_tokens.clear()

            grab_bag.adj_line_to_parse = position_marker.text_to_parse

            (
                indent_level_delta,
                already_adjusted,
                active_container_index,
            ) = ContainerBlockNestedProcessor.__check_for_nested_list_start(
                parser_state,
                nested_container_starts,
                grab_bag,
            )

            assert (
                grab_bag.is_leaf_tokens_empty()
            ), "No leaf tokens are expected at this point."
            adjusted_text_to_parse = ContainerBlockNestedProcessor.__do_nested_cleanup(
                parser_state,
                indent_level_delta,
                already_adjusted,
                active_container_index,
                adjusted_text_to_parse,
                grab_bag,
            )
            # POGGER.debug_with_visible_whitespace("adjusted_text_to_parse>>$>>", adjusted_text_to_parse)

            (
                adjusted_text_to_parse,
                nested_removed_text,
                was_indent_text_added,
                inner_force_list_continuation,
            ) = ContainerBlockNestedProcessor.__check_for_next_container(
                parser_state,
                position_marker,
                nested_container_starts,
                adjusted_text_to_parse,
                grab_bag,
            )
        else:
            (
                nested_removed_text,
                was_indent_text_added,
                inner_force_list_continuation,
            ) = (
                None,
                False,
                False,
            )

        # POGGER.debug_with_visible_whitespace("adjusted_text_to_parse>>$>>", adjusted_text_to_parse)
        grab_bag.line_to_parse = adjusted_text_to_parse
        return (
            nested_removed_text,
            was_indent_text_added,
            inner_force_list_continuation,
        )

    @staticmethod
    def __check_for_next_container(
        parser_state: ParserState,
        position_marker: PositionMarker,
        nested_container_starts: ContainerIndices,
        adjusted_text_to_parse: str,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[str, Optional[str], bool, bool]:
        POGGER.debug("check next container_start>stack>>$", parser_state.token_stack)
        POGGER.debug(
            "check next container_start>tokenized_document>>$",
            parser_state.token_document,
        )

        if (
            nested_container_starts.ulist_index
            or nested_container_starts.olist_index
            or nested_container_starts.block_index
        ):
            POGGER.debug(
                "check next container_start>nested_container",
            )
            (
                adjusted_text_to_parse,
                nested_removed_text,
                was_indent_text_added,
                inner_force_list_continuation,
            ) = ContainerBlockNestedProcessor.__look_for_container_blocks(
                parser_state,
                position_marker,
                grab_bag,
            )
        else:
            (
                nested_removed_text,
                was_indent_text_added,
                inner_force_list_continuation,
            ) = (None, False, False)
        parser_state.set_no_para_start_if_empty()

        return (
            adjusted_text_to_parse,
            nested_removed_text,
            was_indent_text_added,
            inner_force_list_continuation,
        )

    @staticmethod
    def __get_nested_container_starts(
        parser_state: ParserState,
        line_to_parse: str,
        avoid_block_starts: bool,
    ) -> ContainerIndices:
        POGGER.debug(
            "__handle_nested_container_blocks>stack>>:$:<<",
            line_to_parse,
        )

        POGGER.debug("check next container_start>")
        POGGER.debug("check next container_start>stack>>$", parser_state.token_stack)

        _, ex_ws_test = ParserHelper.extract_spaces_verified(line_to_parse, 0)

        whitespace_scan_start_index = 0
        for token_stack_item in parser_state.token_stack:
            if token_stack_item.is_list:
                list_stack_token = cast(ListStackToken, token_stack_item)
                if list_stack_token.ws_before_marker <= len(ex_ws_test):
                    whitespace_scan_start_index = list_stack_token.ws_before_marker

        ws_index, ex_white = ParserHelper.extract_spaces(
            line_to_parse, whitespace_scan_start_index
        )
        ex_whitespace = "" if ex_white is None else ex_white
        after_ws_index: int = (
            whitespace_scan_start_index if ws_index is None else ws_index
        )

        nested_ulist_start, _, _, _ = ListBlockStartsHelper.is_ulist_start(
            parser_state, line_to_parse, after_ws_index, ex_whitespace, False, None
        )
        nested_olist_start, _, _, _ = ListBlockStartsHelper.is_olist_start(
            parser_state, line_to_parse, after_ws_index, ex_whitespace, False, None
        )
        nested_block_start = (
            False
            if avoid_block_starts
            else BlockQuoteCountHelper.is_block_quote_start(
                line_to_parse, after_ws_index, ex_whitespace
            )
        )

        POGGER.debug("check next container_start>stack>>$", parser_state.token_stack)

        return ContainerIndices(
            nested_ulist_start, nested_olist_start, nested_block_start
        )

    @staticmethod
    def __look_for_container_blocks(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[str, Optional[str], bool, bool]:
        """
        Look for container blocks that we can use.
        """
        POGGER.debug("check next container_start>recursing")
        adj_block, position_marker = (
            (
                None
                if grab_bag.end_container_indices.block_index == -1
                else grab_bag.end_container_indices.block_index
            ),
            PositionMarker(position_marker.line_number, -1, grab_bag.adj_line_to_parse),
        )
        new_container_depth = grab_bag.container_depth + 1

        POGGER.debug(
            "position_marker>:$:$:",
            position_marker.index_number,
            position_marker.text_to_parse,
        )
        POGGER.debug("adj_block>:$:", adj_block)
        POGGER.debug("\n\nRECURSING\n")
        # POGGER.debug("parser_state.token_document>$", parser_state.token_document)
        previous_document_length = len(parser_state.token_document)
        (
            produced_inner_tokens,
            adjusted_text_to_parse,
            current_bq_count_increment,
            grab_bag.requeue_line_info,
            grab_bag.did_blank,
            inner_force_list_continuation,
        ) = grab_bag.get_recurse_fn()(
            parser_state,
            position_marker,
            False,
            False,
            grab_bag.parser_properties,
            grab_bag.block_quote_data.current_count,
            new_container_depth,
            adj_block,
            grab_bag.block_quote_data.current_count,
            grab_bag.original_line,
        )
        assert (
            not grab_bag.requeue_line_info
            or not grab_bag.requeue_line_info.lines_to_requeue
        ), "Requeing not supported."
        assert adjusted_text_to_parse is not None, "Adjusted text should be defined."

        POGGER.debug("\nRECURSED\n\n")

        # POGGER.debug("previous_document_length=$", previous_document_length)
        nested_removed_text = (
            ContainerBlockNestedProcessor.__calculate_nested_removed_text(
                parser_state, previous_document_length, grab_bag
            )
        )
        POGGER.debug("check next container_start>stack>>$", parser_state.token_stack)
        POGGER.debug(
            "check next container_start>tokenized_document>>$",
            parser_state.token_document,
        )
        POGGER.debug("check next container_start>line_parse>>$", adjusted_text_to_parse)
        if current_bq_count_increment:
            grab_bag.block_quote_data = BlockQuoteData(
                grab_bag.block_quote_data.current_count + current_bq_count_increment,
                parser_state.count_of_block_quotes_on_stack(),
            )

        POGGER.debug("produced_inner_tokens=$", produced_inner_tokens)
        was_indent_text_added = bool(
            parser_state.token_stack[-1].is_indented_code_block
            and produced_inner_tokens
            and len(produced_inner_tokens) == 1
            and produced_inner_tokens[0].is_text
        )
        POGGER.debug("was_indent_text_added=$", was_indent_text_added)

        # POGGER.debug("parser_state.token_document>$", parser_state.token_document)
        parser_state.token_document.extend(produced_inner_tokens)
        POGGER.debug("parser_state.token_document>$", parser_state.token_document)
        # POGGER.debug("did_process_blank_line>$", did_process_blank_line)

        return (
            adjusted_text_to_parse,
            nested_removed_text,
            was_indent_text_added,
            inner_force_list_continuation,
        )

    @staticmethod
    def __calculate_nested_removed_text(
        parser_state: ParserState,
        previous_document_length: int,
        grab_bag: ContainerGrabBag,
    ) -> Optional[str]:
        nested_removed_text = None
        POGGER.debug("__calculate_nested_removed_text")
        POGGER.debug(
            "__cnrt->token_doc($)",
            ParserHelper.make_value_visible(parser_state.token_document),
        )
        if previous_document_length != len(parser_state.token_document):
            POGGER.debug(
                "\ncheck next container_start>added tokens:$:",
                parser_state.token_document[previous_document_length:],
            )
            if parser_state.token_document[-1].is_block_quote_start:
                block_quote_token = cast(
                    BlockQuoteMarkdownToken, parser_state.token_document[-1]
                )
                nested_removed_text = ContainerBlockNestedProcessor.__calculate_nested_removed_text_block_quote(
                    parser_state, block_quote_token, grab_bag
                )
        if not nested_removed_text:
            POGGER.debug("__cnrt:nested_removed_text:$:", nested_removed_text)
            last_container_index = parser_state.find_last_container_on_stack()
            if (
                last_container_index > 0
                and parser_state.token_stack[last_container_index].is_block_quote
            ):
                block_quote_token = cast(
                    BlockQuoteMarkdownToken,
                    parser_state.token_stack[
                        last_container_index
                    ].matching_markdown_token,
                )
                assert (
                    block_quote_token.bleading_spaces is not None
                ), "Bleading spaces are always defined for block quotes."
                split_spaces = block_quote_token.bleading_spaces.split(
                    ParserHelper.newline_character
                )
                nested_removed_text = str(split_spaces[-1])
        POGGER.debug("__calculate_nested_removed_text<<:$:", nested_removed_text)
        return nested_removed_text

    @staticmethod
    def __calculate_nested_removed_text_block_quote(
        parser_state: ParserState,
        block_quote_token: BlockQuoteMarkdownToken,
        grab_bag: ContainerGrabBag,
    ) -> str:
        assert (
            block_quote_token.bleading_spaces is not None
        ), "Bleading spaces are always defined for block quotes."
        split_spaces = block_quote_token.bleading_spaces.split(
            ParserHelper.newline_character
        )
        nested_removed_text = str(split_spaces[-1])
        POGGER.debug("nested_removed_text:$:", nested_removed_text)
        if (
            len(parser_state.token_document) > 1
            and parser_state.token_document[-2].is_new_list_item
            and parser_state.token_document[-2].line_number
            == parser_state.token_document[-1].line_number
        ):
            column_number_delta = (
                parser_state.token_document[-1].column_number
                - parser_state.token_document[-2].column_number
            )
            block_quote_character_index = nested_removed_text.index(">")
            nested_removed_text_length = len(nested_removed_text)
            POGGER.debug(
                "column_number_delta=$, block_quote_character_index=$, nested_removed_text_length=$",
                column_number_delta,
                block_quote_character_index,
                nested_removed_text_length,
            )
            adjusted_length = (
                parser_state.token_document[-2].column_number
                - 1
                + column_number_delta
                + (nested_removed_text_length - block_quote_character_index)
            )
            POGGER.debug("adjusted_length=$", adjusted_length)
            if adjusted_length != nested_removed_text_length:
                assert (
                    parser_state.original_line_to_parse is not None
                ), "Original line must be defined by this point."
                nested_removed_text = parser_state.original_line_to_parse[
                    :adjusted_length
                ]
                POGGER.debug("previous:$:", nested_removed_text)
                grab_bag.weird_adjusted_text = nested_removed_text
        return nested_removed_text

    @staticmethod
    def __check_for_nested_list_start(
        parser_state: ParserState,
        nested_container_starts: ContainerIndices,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[int, bool, int]:
        active_container_index = max(
            grab_bag.end_container_indices.ulist_index,
            grab_bag.end_container_indices.olist_index,
            grab_bag.end_container_indices.block_index,
        )
        POGGER.debug(
            "check next container_start>max>>$>>bq>>$",
            active_container_index,
            grab_bag.end_container_indices.block_index,
        )
        indent_level_delta, already_adjusted = 0, False
        if (
            grab_bag.end_container_indices.block_index != -1
            and not nested_container_starts.ulist_index
            and not nested_container_starts.olist_index
        ):
            assert (
                active_container_index == grab_bag.end_container_indices.block_index
            ), "Without a list start index, the active container must be a block quote."
            POGGER.debug(
                "parser_state.nested_list_start>>$<<",
                parser_state.nested_list_start,
            )
            POGGER.debug(
                "parser_state.token_document>>$<<", parser_state.token_document
            )
            if parser_state.nested_list_start and grab_bag.adj_line_to_parse.strip(
                Constants.ascii_whitespace
            ):
                (
                    grab_bag.start_index,
                    indent_level,
                    indent_was_adjusted,
                    indent_level_delta,
                ) = ContainerBlockNestedProcessor.__calculate_initial_list_adjustments(
                    parser_state,
                    grab_bag.adj_line_to_parse,
                    grab_bag.end_container_indices,
                )
                already_adjusted = ContainerBlockNestedProcessor.__adjust_line_2(
                    parser_state,
                    indent_level,
                    nested_container_starts,
                    indent_was_adjusted,
                    grab_bag,
                )
        return (
            indent_level_delta,
            already_adjusted,
            active_container_index,
        )

    @staticmethod
    def __adjust_line_2_a(
        parser_state: ParserState, grab_bag: ContainerGrabBag
    ) -> None:
        POGGER.debug("\n\nBOOM\n\n")
        POGGER.debug("parser_state.token_stack=$", parser_state.token_stack)
        POGGER.debug("parser_state.token_document=$", parser_state.token_document)
        POGGER.debug("container_level_tokens=$", grab_bag.container_tokens)

        ending_blank_line_tokens = []
        while parser_state.token_document[-1].is_blank_line:
            ending_blank_line_tokens.append(parser_state.token_document[-1])
            del parser_state.token_document[-1]

        x_tokens, _ = parser_state.close_open_blocks_fn(
            parser_state,
            include_lists=True,
        )
        # test_extra_049l8b - if ct and xt at all end-li and end-bq, empty ct.
        are_x_tokens_all_container_end_tokens = False
        if x_tokens:
            are_x_tokens_all_container_end_tokens = all(
                (x.is_list_end or x.is_block_quote_end) for x in x_tokens
            )
        are_container_tokens_all_container_end_tokens = False
        if grab_bag.container_tokens:
            are_container_tokens_all_container_end_tokens = all(
                (x.is_list_end or x.is_block_quote_end)
                for x in grab_bag.container_tokens
            )

        if (
            are_x_tokens_all_container_end_tokens
            and are_container_tokens_all_container_end_tokens
        ):
            parser_state.token_document.extend(grab_bag.container_tokens)
            grab_bag.clear_container_tokens()
        parser_state.token_document.extend(x_tokens)

        parser_state.token_document.extend(ending_blank_line_tokens)

        POGGER.debug("parser_state.token_stack=$", parser_state.token_stack)
        POGGER.debug("parser_state.token_document=$", parser_state.token_document)
        POGGER.debug("container_level_tokens=$", grab_bag.container_tokens)

    @staticmethod
    def __adjust_line_2(
        parser_state: ParserState,
        indent_level: int,
        nested_container_starts: ContainerIndices,
        indent_was_adjusted: bool,
        grab_bag: ContainerGrabBag,
    ) -> bool:
        """
        Handle the case where the document currently ends with blank lines, but there
        are end list and end block quote tokens that need to be processed ahead of the
        blank line.  This occurs because the blank line forced the end container tokens
        to occur, and just needs to be rearranged to make sense.
        """

        if (
            parser_state.token_document[-1].is_blank_line
            and (grab_bag.end_container_indices.block_index + grab_bag.start_index)
            < indent_level
        ):
            ContainerBlockNestedProcessor.__adjust_line_2_a(parser_state, grab_bag)
        elif (
            not nested_container_starts.block_index
            and grab_bag.adj_line_to_parse
            and grab_bag.adj_line_to_parse[0] == ParserHelper.space_character
            and indent_was_adjusted
            and parser_state.nested_list_start
        ):
            assert (
                parser_state.nested_list_start is not None
            ), "The nested list start cannot be None."
        return False

    # pylint: disable=too-many-locals
    @staticmethod
    def __calculate_initial_list_adjustments(
        parser_state: ParserState,
        adj_line_to_parse: str,
        end_container_indices: ContainerIndices,
    ) -> Tuple[int, int, bool, int]:
        start_index, _ = ParserHelper.extract_spaces_verified(adj_line_to_parse, 0)
        POGGER.debug("start_index>>$<<", start_index)

        assert (
            parser_state.nested_list_start is not None
        ), "The nested list start cannot be None."
        assert (
            parser_state.nested_list_start.matching_markdown_token is not None
        ), "List stack starts always have matching markdown tokens."
        POGGER.debug(
            "parser_state.nested_list_start.matching_markdown_token>>$<<",
            parser_state.nested_list_start.matching_markdown_token,
        )
        mmt_list = cast(
            ListStartMarkdownToken,
            parser_state.nested_list_start.matching_markdown_token,
        )
        list_start_token_index = parser_state.token_document.index(mmt_list)
        POGGER.debug(
            "list_start_token_index>>$<<",
            list_start_token_index,
        )
        if list_start_token_index < (len(parser_state.token_document) - 1):
            token_after_list_start = parser_state.token_document[
                list_start_token_index + 1
            ]
            POGGER.debug(
                "token_after_list_start>>$<<",
                token_after_list_start,
            )
            if token_after_list_start.is_setext_heading:
                setext_token_after_list_start = cast(
                    SetextHeadingMarkdownToken, token_after_list_start
                )
                line_number = setext_token_after_list_start.original_line_number
                column_number = setext_token_after_list_start.original_column_number
            else:
                line_number = token_after_list_start.line_number
                column_number = token_after_list_start.column_number
            assert (
                mmt_list.line_number == line_number
            ), "Token after the list start must have the same line number as the list start."

            total_list_start = (
                mmt_list.list_start_content + mmt_list.list_start_sequence
            )
            column_number_delta = column_number - mmt_list.column_number
            if token_after_list_start.is_blank_line and column_number_delta == len(
                total_list_start
            ):
                column_number_delta += 1
        else:
            column_number_delta = 0
        POGGER.debug(
            "column_number_delta>>$<<",
            column_number_delta,
        )
        indent_level_delta, indent_level, adjusted_indent_level = (
            0,
            parser_state.nested_list_start.indent_level,
            column_number_delta + end_container_indices.block_index,
        )
        POGGER.debug(
            "adjusted_indent_level>>$<<  indent_level>$",
            adjusted_indent_level,
            indent_level,
        )
        indent_was_adjusted = indent_level != adjusted_indent_level
        if indent_level > adjusted_indent_level:
            indent_level_delta = indent_level - adjusted_indent_level
        indent_level = column_number_delta + end_container_indices.block_index

        return start_index, indent_level, indent_was_adjusted, indent_level_delta

    # pylint: enable=too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __do_nested_cleanup(
        parser_state: ParserState,
        delta: int,
        already_adjusted: bool,
        active_container_index: int,
        adjusted_text_to_parse: str,
        grab_bag: ContainerGrabBag,
    ) -> str:
        if delta or already_adjusted:
            adjusted_text_to_parse = grab_bag.adj_line_to_parse
        else:
            POGGER.debug("active_container_index<<$<<", active_container_index)
            adjustment_filler = ParserHelper.repeat_string(
                ParserHelper.space_character, active_container_index
            )
            grab_bag.adj_line_to_parse = (
                f"{adjustment_filler}{grab_bag.adj_line_to_parse}"
            )

        parser_state.token_document.extend(grab_bag.container_tokens)
        grab_bag.clear_container_tokens()
        return adjusted_text_to_parse

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods



"""
Module to implement a plugin that looks for trailing spaces in the files.
"""

from typing import List, Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.container_markdown_token import ContainerMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken


# pylint: disable=too-many-instance-attributes
class RuleMd009(RulePlugin):
    """
    Class to implement a plugin that looks for trailing spaces in the files.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__leaf_tokens: List[MarkdownToken] = []
        self.__leaf_owner_tokens: List[Optional[MarkdownToken]] = []
        self.__line_index = 0
        self.__leaf_token_index = 0
        self.__inline_token_index = 0
        self.__container_token_stack: List[ContainerMarkdownToken] = []
        self.__break_spaces = 0
        self.__strict_mode = False
        self.__list_item_empty_lines_mode = False
        self.__last_line: Optional[str] = None

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="no-trailing-spaces",
            plugin_id="MD009",
            plugin_enabled_by_default=True,
            plugin_description="Trailing spaces",
            plugin_version="0.6.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md009.md",
            plugin_configuration="br_spaces,list_item_empty_lines,strict",
            plugin_supports_fix=True,
            plugin_fix_level=0,
        )

    @classmethod
    def __validate_break_spaces(cls, found_value: int) -> None:
        if found_value < 0:
            raise ValueError("Allowable values are greater than or equal to 0.")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__break_spaces = (
            self.plugin_configuration.get_integer_property_with_default(
                "br_spaces",
                2,
                valid_value_fn=self.__validate_break_spaces,
            )
        )
        if self.__break_spaces < 2:
            self.__break_spaces = 0

        self.__strict_mode = (
            self.plugin_configuration.get_boolean_property_with_default(
                "strict",
                False,
            )
        )

        self.__list_item_empty_lines_mode = (
            self.plugin_configuration.get_boolean_property_with_default(
                "list_item_empty_lines",
                False,
            )
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("br_spaces", self.__break_spaces),
            QueryConfigItem("strict", self.__strict_mode),
            QueryConfigItem("list_item_empty_lines", self.__list_item_empty_lines_mode),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__leaf_tokens = []
        self.__leaf_owner_tokens = []
        self.__line_index = 1
        self.__leaf_token_index = 0
        self.__inline_token_index = 0
        self.__container_token_stack = []
        self.__last_line = None

    def next_line(self, context: PluginScanContext, line: str) -> None:
        """
        Event that a new line is being processed.
        """
        if (
            self.__leaf_token_index + 1 < len(self.__leaf_tokens)
            and self.__line_index
            == self.__leaf_tokens[self.__leaf_token_index + 1].line_number
            and self.__leaf_tokens[self.__leaf_token_index + 1].is_leaf
        ):
            self.__leaf_token_index = self.__inline_token_index + 1
            self.__inline_token_index = self.__leaf_token_index
        if (
            self.__inline_token_index + 1 < len(self.__leaf_tokens)
            and self.__line_index
            == self.__leaf_tokens[self.__inline_token_index + 1].line_number
        ):
            self.__inline_token_index += 1

        if (
            self.__leaf_token_index < len(self.__leaf_tokens)
            and not self.__leaf_tokens[self.__leaf_token_index].is_code_block
            and line
            and line[-1] == " "
        ):
            self.__next_line_check_for_error(line, context)
        self.__line_index += 1
        self.__last_line = line.strip(" \t")

    def __next_line_check_for_error(
        self, line: str, context: PluginScanContext
    ) -> None:
        (
            first_non_whitespace_index,
            extracted_whitespace,
        ) = ParserHelper.extract_spaces_from_end(line)
        extracted_whitespace_length = len(extracted_whitespace)

        is_within_list = False
        new_list_indent = -1
        expected_list_indent = -1
        if self.__list_item_empty_lines_mode and first_non_whitespace_index == 0:
            leaf_token = self.__leaf_owner_tokens[self.__leaf_token_index]
            if leaf_token is not None:
                is_within_list = True
                list_owner_token = cast(ListStartMarkdownToken, leaf_token)
                expected_list_indent = (
                    list_owner_token.indent_level + self.__break_spaces
                )
                if extracted_whitespace_length != expected_list_indent:
                    new_list_indent = expected_list_indent

        if (
            not is_within_list
            and (
                extracted_whitespace_length != self.__break_spaces or self.__strict_mode
            )
            or new_list_indent != -1
        ):
            if context.in_fix_mode:
                self.__report_fix(
                    context,
                    line,
                    new_list_indent,
                    extracted_whitespace_length,
                    first_non_whitespace_index,
                )
            else:
                self.__report_error(
                    context,
                    extracted_whitespace_length,
                    first_non_whitespace_index,
                    expected_list_indent,
                )

    # pylint: disable=too-many-arguments
    def __report_fix(
        self,
        context: PluginScanContext,
        line: str,
        new_list_indent: int,
        extracted_whitespace_length: int,
        first_non_whitespace_index: int,
    ) -> None:
        if new_list_indent != -1:
            line = " " * new_list_indent
        else:
            this_leaf_token = self.__leaf_tokens[self.__leaf_token_index]
            line = (
                line[:first_non_whitespace_index]
                if (
                    extracted_whitespace_length < self.__break_spaces
                    or this_leaf_token.is_atx_heading
                    or self.__strict_mode
                )
                else line[: first_non_whitespace_index + self.__break_spaces]
            )
        context.set_current_fix_line(line)

    # pylint: enable=too-many-arguments

    def __report_error(
        self,
        context: PluginScanContext,
        extracted_whitespace_length: int,
        first_non_whitespace_index: int,
        expected_list_indent: int,
    ) -> None:
        if expected_list_indent != -1:
            extra_error_information = str(expected_list_indent)
        elif self.__strict_mode or self.__break_spaces < 2:
            extra_error_information = "0"
        else:
            extra_error_information = f"0 or {self.__break_spaces}"
        extra_error_information = f"Expected: {extra_error_information}; Actual: {extracted_whitespace_length}"
        override_is_error_token_prefaced_by_blank_line = (
            self.__last_line is not None and not self.__last_line
        )
        self.report_next_line_error(
            context,
            first_non_whitespace_index + 1,
            extra_error_information=extra_error_information,
            override_is_error_token_prefaced_by_blank_line=override_is_error_token_prefaced_by_blank_line,
        )

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        _ = context

        if token.is_block_quote_start or token.is_list_start:
            container_token = cast(ContainerMarkdownToken, token)
            self.__container_token_stack.append(container_token)
        elif token.is_block_quote_end or token.is_list_end:
            del self.__container_token_stack[-1]
        elif token.is_blank_line or token.is_leaf or token.is_inline_hard_break:
            self.__leaf_tokens.append(token)
            if self.__container_token_stack:
                self.__leaf_owner_tokens.append(self.__container_token_stack[-1])
            else:
                self.__leaf_owner_tokens.append(None)


# pylint: enable=too-many-instance-attributes



"""
Module to provide context when reporting any errors.
"""

from __future__ import annotations

from io import TextIOWrapper
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple, Union

from typing_extensions import override

from pymarkdown.plugin_manager.bad_plugin_error import BadPluginError
from pymarkdown.plugin_manager.fix_line_record import FixLineRecord
from pymarkdown.plugin_manager.fix_token_record import FixTokenRecord
from pymarkdown.plugin_manager.plugin_modify_context import PluginModifyContext
from pymarkdown.plugin_manager.plugin_scan_failure import PluginScanFailure
from pymarkdown.plugin_manager.replace_tokens_record import ReplaceTokensRecord
from pymarkdown.tokens.markdown_token import MarkdownToken

if TYPE_CHECKING:  # pragma: no cover
    from pymarkdown.plugin_manager.plugin_manager import PluginManager


# pylint: disable=too-many-instance-attributes
class PluginScanContext(PluginModifyContext):
    """
    Class to provide context when reporting any errors.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        owning_manager: PluginManager,
        scan_file: str,
        actual_tokens: List[MarkdownToken],
        fix_mode: bool,
        file_output: Optional[TextIOWrapper],
        fix_token_map: Optional[Dict[MarkdownToken, List[FixTokenRecord]]],
        replace_tokens_list: Optional[List[ReplaceTokensRecord]],
    ):
        self.owning_manager, self.scan_file, self.line_number = (
            owning_manager,
            scan_file,
            0,
        )
        self.__reported: List[PluginScanFailure] = []
        self.__in_fix_mode = fix_mode
        self.__current_fix_line: Optional[str] = None
        self.__last_line_fixed: Optional[str] = None
        self.__line_change_record: List[FixLineRecord] = []
        self.__file_output = file_output
        self.__fix_token_map = fix_token_map
        self.__replace_token_list = replace_tokens_list
        self.__actual_tokens = actual_tokens

    # pylint: enable=too-many-arguments

    def register_replace_tokens_request(
        self,
        plugin_id: str,
        start_token: MarkdownToken,
        end_token: MarkdownToken,
        replacement_tokens: List[MarkdownToken],
    ) -> None:
        """
        Register a sequence of tokens and what to replace them with.
        """
        assert self.__replace_token_list is not None
        new_record = ReplaceTokensRecord(
            plugin_id, start_token, end_token, replacement_tokens
        )
        self.__replace_token_list.append(new_record)

    # pylint: disable=too-many-arguments
    def register_fix_token_request(
        self,
        token: MarkdownToken,
        plugin_id: str,
        plugin_action: str,
        field_name: str,
        field_value: Union[str, int],
    ) -> None:
        """
        Register a token to fix and how to fix it.
        """
        if self.__fix_token_map is None:
            return

        new_record = FixTokenRecord(
            token, plugin_id, plugin_action, field_name, field_value
        )
        if token not in self.__fix_token_map:
            existing_records: List[FixTokenRecord] = []
            self.__fix_token_map[token] = existing_records
        else:
            existing_records = self.__fix_token_map[token]
        existing_records.append(new_record)

    # pylint: enable=too-many-arguments

    def get_fix_token_map(self) -> Dict[MarkdownToken, List[FixTokenRecord]]:
        """
        Get the current map of tokens to fix and how to fix them.
        """
        assert self.__fix_token_map is not None
        return self.__fix_token_map

    def get_replace_tokens_list(self) -> List[ReplaceTokensRecord]:
        """
        Get the current list of replacement records/tokens.
        """
        assert self.__replace_token_list is not None
        return self.__replace_token_list

    @property
    @override
    def is_during_line_pass(self) -> bool:
        return self.__file_output is not None

    @property
    @override
    def in_fix_mode(self) -> bool:
        """
        Report on whether the application is in fix mode.
        """
        return self.__in_fix_mode

    @property
    def last_line_fixed(self) -> Optional[str]:
        """
        If in fix mode and processing lines, contains a transformed line.
        """
        return self.__last_line_fixed

    def set_last_line_fixed(self, new_line: Optional[str]) -> None:
        """
        While fixing a document line-by-line, set the current line's content.
        """
        self.__last_line_fixed = new_line

    @property
    def current_fix_line(self) -> Optional[str]:
        """
        If in fix mode and processing lines, contains a transformed line.
        """
        return self.__current_fix_line

    def set_current_fix_line(self, new_line: Optional[str]) -> None:
        """
        While fixing a document line-by-line, set the current line's content.
        """
        self.__current_fix_line = new_line

    @property
    def file_output(self) -> TextIOWrapper:
        """
        File object to use for writing fix information.
        """
        assert self.__file_output
        return self.__file_output

    def add_fix_line_record(self, change_record: FixLineRecord) -> None:
        """
        Add a record regarding a line fix that was made.
        """
        self.__line_change_record.append(change_record)

    @property
    def fix_line_records(self) -> List[FixLineRecord]:
        """
        Get any records detailing line fixes.
        """
        return self.__line_change_record

    # pylint: disable=too-many-arguments
    def add_triggered_rule(
        self,
        scan_file: str,
        line_number: int,
        column_number: int,
        rule_id: str,
        rule_name: str,
        rule_description: str,
        extra_error_information: Optional[str],
        does_support_fix: bool,
        error_token: Optional[MarkdownToken] = None,
        override_is_error_token_prefaced_by_blank_line: Optional[bool] = None,
    ) -> None:
        """
        Add the triggering information for a rule.
        """
        if self.in_fix_mode:
            if does_support_fix:
                raise BadPluginError(
                    formatted_message=f"Plugin {rule_id}({rule_name}) reported a triggered rule while in fix mode."
                )
            return

        if override_is_error_token_prefaced_by_blank_line is not None:
            is_error_token_prefaced_by_blank_line = (
                override_is_error_token_prefaced_by_blank_line
            )
        else:
            is_error_token_prefaced_by_blank_line = self.__calc_x(error_token)

        new_entry = PluginScanFailure(
            scan_file,
            line_number,
            column_number,
            rule_id,
            rule_name,
            rule_description,
            extra_error_information,
            is_error_token_prefaced_by_blank_line,
        )
        self.__reported.append(new_entry)

    # pylint: enable=too-many-arguments

    def __calc_x_rewind_if_inline(
        self, index_to_check: int, current_token: MarkdownToken, dd: bool
    ) -> Tuple[int, MarkdownToken, bool, bool]:
        # For an inline, rewind back to the last text token.  If the text token was
        # not on the same line as the inline token, by definition there is text on
        # the line, so break out of this check.
        #
        # Note that in special cases, such as a link being at the start of a parapgrah,
        # we may not have any text element before the paragraph element.
        do_break = False
        if (
            not current_token.is_leaf
            and not current_token.is_container
            and not current_token.is_text
        ):
            while (
                index_to_check >= 0
                and not self.__actual_tokens[index_to_check].is_text
                and not self.__actual_tokens[index_to_check].is_paragraph
            ):
                index_to_check -= 1
            assert index_to_check >= 0
            if (
                current_token.line_number
                != self.__actual_tokens[index_to_check].line_number
            ):
                do_break = True
            else:
                current_token = self.__actual_tokens[index_to_check]
                dd = True
        return index_to_check, current_token, dd, do_break

    def __calc_x_rewind_if_text(
        self, index_to_check: int, current_token: MarkdownToken, dd: bool
    ) -> Tuple[int, MarkdownToken, bool]:
        if current_token.is_text:
            # Text blocks can only occur within leaf elements, so work backwards until
            # we hit a leaf element.
            while (
                index_to_check >= 0 and not self.__actual_tokens[index_to_check].is_leaf
            ):
                index_to_check -= 1
            if index_to_check > 0:
                index_to_check -= 1
            current_token = self.__actual_tokens[index_to_check]
            dd = True
        elif current_token.is_paragraph:
            if index_to_check > 0:
                index_to_check -= 1
            current_token = self.__actual_tokens[index_to_check]
        return index_to_check, current_token, dd

    def __calc_x(self, error_token: Optional[MarkdownToken]) -> bool:
        if error_token is None:
            return False

        is_error_token_prefaced_by_blank_line = False
        for i, j in enumerate(self.__actual_tokens):  # pragma: no cover
            if not (
                j.line_number == error_token.line_number
                and j.column_number == error_token.column_number
                and j.token_name == error_token.token_name
            ):
                continue

            index_to_check = i
            current_token = error_token
            dd = False

            index_to_check, current_token, dd, do_break = (
                self.__calc_x_rewind_if_inline(index_to_check, current_token, dd)
            )
            if do_break:
                break

            index_to_check, current_token, dd = self.__calc_x_rewind_if_text(
                index_to_check, current_token, dd
            )

            # Lists can have trailing
            if current_token.is_list_start:
                while (
                    index_to_check >= 0
                    and self.__actual_tokens[index_to_check - 1].is_list_end
                ):
                    index_to_check -= 1
                    # should we be doing this more generally? i.e. list closing and opening a fcb... ?
                    # assert False  -- check to see if we need is_list_end or is_block_quote_end
                if index_to_check > 0:
                    index_to_check -= 1
                current_token = self.__actual_tokens[index_to_check]
                dd = True
            assert index_to_check >= 0
            if not dd and index_to_check > 0:
                index_to_check -= 1
            is_error_token_prefaced_by_blank_line = self.__actual_tokens[
                index_to_check
            ].is_blank_line
            break
        return is_error_token_prefaced_by_blank_line

    def report_on_triggered_rules(self) -> None:
        """
        Report on the various points where rules were triggered,
        in sorted order.
        """
        for next_entry in sorted(self.__reported):
            self.owning_manager.log_scan_failure(next_entry)
        self.__reported.clear()

    def get_triggered_rules(self) -> Set[str]:
        """
        Get information on any rules that were triggered.
        """
        return {next_entry.rule_id.lower() for next_entry in self.__reported}

    def is_pragma_on_line(self, line_number: int) -> bool:
        """
        Determine if there is a pragma on the specified line.
        """
        return self.owning_manager.is_pragma_on_line(line_number)

    def calc_pragma_offset(self, token: MarkdownToken, line_number_delta: int) -> int:
        """
        Calculate the pragma offset for a given token and line number delta.
        """
        pragma_offset = 0
        for line_index in range(line_number_delta):
            modified_line_number = token.line_number + line_index + 1 + pragma_offset
            if self.owning_manager.is_pragma_on_line(modified_line_number):
                pragma_offset += 1
        return pragma_offset


# pylint: enable=too-many-instance-attributes



"""
Module to provide for a transformation from markdown tokens to html for GFM.
"""

import logging
from typing import List

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_gfm.transform_to_gfm_token_handlers import (
    TransformToGfmTokenHandlers,
)

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods
class TransformToGfm:
    """
    Class to provide for a transformation from markdown tokens to html for GFM.
    """

    add_trailing_text_tokens = [
        "<hr />",
        "<p>",
        "<h1>",
        "<h2>",
        "<h3>",
        "<h4>",
        "<h5>",
        "<h6>",
        "<pre>",
        "<ul>",
        "<ol>",
        '<ol start="',
    ]

    def __init__(self) -> None:
        self.__token_handlers = TransformToGfmTokenHandlers()

    def transform(self, actual_tokens: List[MarkdownToken]) -> str:
        """
        Transform the tokens into html.
        """
        POGGER.debug("\n\n---\n")
        transform_state, output_html, actual_tokens_size = (
            TransformState(actual_tokens),
            "",
            len(actual_tokens),
        )

        # This is the easiest way to finish covering the missing items.
        assert (
            transform_state.next_token is None
            and not transform_state.is_in_fenced_code_block
        ), "Initial state must be set properly."

        for next_token in transform_state.actual_tokens:
            output_html = self.__token_handlers.apply_transformation(
                transform_state,
                actual_tokens,
                actual_tokens_size,
                next_token,
                output_html,
            )

            POGGER.debug("======")
            POGGER.debug(
                "add_trailing_text-->$<--",
                transform_state.add_trailing_text,
            )
            POGGER.debug("add_leading_text -->$<--", transform_state.add_leading_text)
            POGGER.debug("output_html    -->$<--", output_html)

            if transform_state.add_trailing_text:
                output_html = self.__apply_trailing_text(output_html, transform_state)
                POGGER.debug("output_html    -->$<--", output_html)

            if transform_state.add_leading_text:
                output_html = self.__apply_leading_text(output_html, transform_state)
                POGGER.debug("output_html    -->$<--", output_html)

            POGGER.debug("------")
            POGGER.debug("next_token     -->$<--", next_token)
            POGGER.debug("output_html    -->$<--", output_html)
            POGGER.debug("transform_stack-->$<--", transform_state.transform_stack)

            transform_state.last_token = next_token
            transform_state.actual_token_index += 1
        if output_html and output_html[-1] == ParserHelper.newline_character:
            output_html = output_html[:-1]
        POGGER.debug("output_html    -->$<--", output_html)
        return output_html

    @classmethod
    def __apply_trailing_text(
        cls, output_html: str, transform_state: TransformState
    ) -> str:
        """
        Apply any trailing text to the output.
        """
        POGGER.debug("__apply_trailing_text>:$:<", output_html)
        stack_text = transform_state.transform_stack.pop()
        trailing_part = [stack_text]
        for next_token_to_test in TransformToGfm.add_trailing_text_tokens:
            if output_html.startswith(next_token_to_test):
                trailing_part.append(ParserHelper.newline_character)
                break

        POGGER.debug("trailing_part>:$:<", trailing_part)
        if trailing_part[-1].endswith("<li>") and output_html.startswith(
            "<blockquote>"
        ):
            trailing_part.append(ParserHelper.newline_character)
        trailing_part.append(output_html)
        POGGER.debug("trailing_part>:$:<", trailing_part)
        if output_html.endswith("</ul>") or output_html.endswith("</ol>"):
            trailing_part.append(ParserHelper.newline_character)
        assert (
            transform_state.add_trailing_text is not None
        ), "Trailing text must be defined by now."
        trailing_part.append(transform_state.add_trailing_text)
        combined_text = "".join(trailing_part)
        POGGER.debug("__apply_trailing_text>:$:<", combined_text)
        return combined_text

    @classmethod
    def __apply_leading_text(
        cls, output_html: str, transform_state: TransformState
    ) -> str:
        """
        Apply any leading text to the output.
        """

        output_html = (
            f"{output_html}{ParserHelper.newline_character}{transform_state.add_leading_text}"
            if output_html and output_html[-1] != ParserHelper.newline_character
            else f"{output_html}{transform_state.add_leading_text}"
        )
        transform_state.transform_stack.append(output_html)
        return ""


# pylint: enable=too-few-public-methods



"""
Module to provide helper functions for parsing.
"""

import logging
from typing import Any, List, Optional, Tuple

from pymarkdown.general.constants import Constants

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
class ParserHelper:
    """
    Class to provide helper functions for parsing.
    """

    __backspace_character = "\b"
    __alert_character = "\a"
    whitespace_split_character = "\x02"
    replace_noop_character = "\x03"
    escape_character = "\x05"

    backslash_character = "\\"
    newline_character = "\n"
    tab_character = "\t"
    space_character = " "

    all_escape_characters = (
        f"{__backspace_character}{__alert_character}{whitespace_split_character}"
        + f"{replace_noop_character}{escape_character}"
    )

    backslash_escape_sequence = f"{backslash_character}{__backspace_character}"

    __normal_whitespace = f" {tab_character}"

    @staticmethod
    def is_character_at_index(
        source_string: str, index_in_string: int, valid_character: str
    ) -> bool:
        """
        Determine if the specified character is at a valid location and is the
        specified valid character.
        """

        return (
            0 <= index_in_string < len(source_string)
            and source_string[index_in_string] == valid_character
        )

    @staticmethod
    def are_characters_at_index(
        source_string: str, index_in_string: int, string_to_match: str
    ) -> bool:
        """
        Determine if the specified character is at a valid location and is the
        specified valid character.
        """

        test_index = index_in_string + len(string_to_match)
        return (
            index_in_string >= 0
            and test_index <= len(source_string)
            and source_string[index_in_string:test_index] == string_to_match
        )

    @staticmethod
    def is_character_at_index_not(
        source_string: str, index_in_string: int, valid_character: str
    ) -> bool:
        """
        Determine if the specified character is at a valid location and is
        not the specified valid character.
        """

        return (
            0 <= index_in_string < len(source_string)
            and source_string[index_in_string] != valid_character
        )

    @staticmethod
    def is_character_at_index_whitespace(
        source_string: str, index_in_string: int
    ) -> bool:
        """
        Determine if the specified character is valid and a whitespace character.
        """

        return (
            0 <= index_in_string < len(source_string)
            and source_string[index_in_string] in ParserHelper.__normal_whitespace
        )

    @staticmethod
    def is_character_at_index_not_whitespace(
        source_string: str, index_in_string: int
    ) -> bool:
        """
        Determine if the specified character is valid and not a whitespace character.
        """

        return (
            0 <= index_in_string < len(source_string)
            and source_string[index_in_string] not in ParserHelper.__normal_whitespace
        )

    @staticmethod
    def is_character_at_index_one_of(
        source_string: str, index_in_string: int, valid_characters: str
    ) -> bool:
        """
        Determine if the specified character is at a valid location and is one
        of the specified valid characters.
        """

        return (
            0 <= index_in_string < len(source_string)
            and source_string[index_in_string] in valid_characters
        )

    @staticmethod
    def is_character_at_index_not_one_of(
        source_string: str, index_in_string: int, valid_characters: str
    ) -> bool:
        """
        Determine if the specified character is at a valid location and is not one
        of the specified valid characters.
        """

        return (
            0 <= index_in_string < len(source_string)
            and source_string[index_in_string] not in valid_characters
        )

    @staticmethod
    def extract_spaces(
        source_string: str, start_index: int
    ) -> Tuple[Optional[int], Optional[str]]:
        """
        From the start_index, continue extracting whitespace while we have it.

        Returns a tuple with the index of the first non-whitespace character and any extracted
        whitespace
        OR
        a tuple of None values if the start_index is not within the source_string.
        """

        if not 0 <= start_index <= len(source_string):
            return None, None

        index = start_index
        while ParserHelper.is_character_at_index_whitespace(source_string, index):
            index += 1

        return index, source_string[start_index:index]

    @staticmethod
    def extract_spaces_verified(
        source_string: str, start_index: int
    ) -> Tuple[int, str]:
        """
        From the start_index, continue extracting whitespace while we have it.

        Returns a tuple with the index of the first non-whitespace character and any extracted
        whitespace.  If start_index is not within the source_string,
        this function will throw an AssertError.
        """
        index, extracted_string = ParserHelper.extract_spaces(
            source_string, start_index
        )
        assert index is not None
        assert extracted_string is not None
        return index, extracted_string

    @staticmethod
    def extract_ascii_whitespace(
        source_string: str, start_index: int
    ) -> Tuple[Optional[int], Optional[str]]:
        """
        From the start_index, continue extracting whitespace while we have it.

        Returns a tuple with the index of the first non-whitespace character and
        any extracted whitespace
        OR
        a tuple of None values if the start_index is not within the source_string.
        """

        if not 0 <= start_index <= len(source_string):
            return None, None

        index = start_index
        while ParserHelper.is_character_at_index_one_of(
            source_string, index, Constants.ascii_whitespace
        ):
            index += 1

        return index, source_string[start_index:index]

    @staticmethod
    def extract_ascii_whitespace_verified(
        source_string: str, start_index: int
    ) -> Tuple[int, str]:
        """
        From the start_index, continue extracting whitespace while we have it.

        Returns a tuple with the index of the first non-whitespace character and
        any extracted whitespace.  If start_index is not within the source_string,
        this function will throw an AssertError.
        """
        index, extracted_string = ParserHelper.extract_ascii_whitespace(
            source_string, start_index
        )
        assert index is not None
        assert extracted_string is not None
        return index, extracted_string

    @staticmethod
    def extract_spaces_from_end(
        source_string: str, start_index: Optional[int] = None
    ) -> Tuple[int, str]:
        """
        From the end of the string, continue extracting whitespace while we have it.

        Returns the index of the last non-whitespace character and any extracted whitespace
        in a tuple.
        """
        if not source_string:
            return 0, ""

        index = (
            (start_index - 1) if start_index is not None else (len(source_string) - 1)
        )
        while ParserHelper.is_character_at_index_whitespace(source_string, index):
            index -= 1

        return index + 1, source_string[index + 1 :]

    @staticmethod
    def extract_until_spaces(
        source_string: str, start_index: int
    ) -> Tuple[Optional[int], Optional[str]]:
        """
        From the start_index, continue extracting until we hit whitespace.

        Returns a tuple with the index of the first whitespace character and any
        extracted text
        OR
        a Tuple of None values if the start_index is not within the source_string.
        """

        if not 0 <= start_index <= len(source_string):
            return None, None

        index = start_index
        while ParserHelper.is_character_at_index_not_whitespace(source_string, index):
            index += 1

        return index, source_string[start_index:index]

    @staticmethod
    def extract_until_spaces_verified(
        source_string: str, start_index: int
    ) -> Tuple[int, str]:
        """
        From the start_index, continue extracting until we hit whitespace.

        Returns a tuple with the index of the first whitespace character and any
        extracted text. If start_index is not within the source_string, this
        function will throw an AssertError.
        """

        index, extracted_string = ParserHelper.extract_until_spaces(
            source_string, start_index
        )
        assert index is not None
        assert extracted_string is not None
        return index, extracted_string

    @staticmethod
    def collect_while_character(
        source_string: str, start_index: int, match_character: str
    ) -> Tuple[Optional[int], Optional[int]]:
        """
        Collect a sequence of the same character from a given starting point in a string.

        Returns a tuple with the number of characters collected and the index of the first
        non-matching character and any extracted text
        OR
        a tuple of None values if the start_index is not within the source_string.
        """

        source_string_size = len(source_string)

        if not 0 <= start_index <= source_string_size:
            return None, None

        index = start_index
        while index < source_string_size and source_string[index] == match_character:
            index += 1
        return index - start_index, index

    @staticmethod
    def collect_while_character_verified(
        source_string: str, start_index: int, match_character: str
    ) -> Tuple[int, int]:
        """
        Collect a sequence of the same character from a given starting point in a string.

        Returns a tuple with the number of characters collected and the index of the first
        non-matching character and any extracted text. If start_index is not within the
        source_string, this function will throw an AssertError.
        """
        index, extracted_string = ParserHelper.collect_while_character(
            source_string, start_index, match_character
        )
        assert index is not None
        assert extracted_string is not None
        return index, extracted_string

    @staticmethod
    def collect_backwards_while_spaces(
        source_string: str, end_index: int
    ) -> Tuple[Optional[int], Optional[int]]:
        """
        Collect from a given starting point in a string going backwards
        towards the start of the string while the character is a space character
        or a tab character.

        Returns a tuple with the number of characters collected and the index of
        he first non-matching character and any extracted text
        OR
        a tuple of None values if the end_index is not within the source_string.
        """
        return ParserHelper.collect_backwards_while_one_of_characters(
            source_string, end_index, ParserHelper.__normal_whitespace
        )

    @staticmethod
    def collect_backwards_while_spaces_verified(
        source_string: str, end_index: int
    ) -> Tuple[int, int]:
        """
        Collect from a given starting point in a string going backwards
        towards the start of the string while the character is a space character
        or a tab character.

        Returns a tuple with the number of characters collected and the index of
        he first non-matching character and any extracted text.  If end_index is
        not within the source_string, this function will throw an AssertError.
        """
        index, extracted_string = ParserHelper.collect_backwards_while_spaces(
            source_string, end_index
        )
        assert index is not None
        assert extracted_string is not None
        return index, extracted_string

    @staticmethod
    def collect_backwards_while_character(
        source_string: str, end_index: int, match_character: str
    ) -> Tuple[Optional[int], Optional[int]]:
        """
        Collect a sequence of the same character from a given starting point in a
        string going backwards towards the start of the string.

        Returns a tuple with the number of characters collected and the index of
        the first non-matching character and any extracted text
        OR
        a tuple of None values if the end_index is not within the source_string.
        """

        source_string_size = len(source_string)
        if not -1 <= end_index <= source_string_size:
            return None, None
        if end_index == -1:
            end_index = source_string_size

        index = end_index
        while index and source_string[index - 1] == match_character:
            index -= 1
        return end_index - index, index

    @staticmethod
    def collect_backwards_while_character_verified(
        source_string: str, end_index: int, match_character: str
    ) -> Tuple[int, int]:
        """
        Collect a sequence of the same character from a given starting point in a
        string going backwards towards the start of the string.

        Returns a tuple with the number of characters collected and the index of
        the first non-matching character and any extracted text. If end_index is
        not within the source_string, this function will throw an AssertError.
        """
        index, extracted_string = ParserHelper.collect_backwards_while_character(
            source_string, end_index, match_character
        )
        assert index is not None
        assert extracted_string is not None
        return index, extracted_string

    @staticmethod
    def collect_backwards_while_one_of_characters(
        source_string: str, end_index: int, match_characters: str
    ) -> Tuple[Optional[int], Optional[int]]:
        """
        Collect a sequence of the same character from a given starting point in
        a string going backwards towards the start of the string.

        Returns a tuple with the number of characters collected and the index of
        the first non-matching character and any extracted text
        OR
        a tuple of None values if the end_index is not within the source_string
        or the value of -1.
        """

        source_string_size = len(source_string)
        if not -1 <= end_index <= source_string_size:
            return None, None
        if end_index == -1:
            end_index = source_string_size

        index = end_index
        while index and source_string[index - 1] in match_characters:
            index -= 1
        return end_index - index, index

    @staticmethod
    def collect_backwards_while_one_of_characters_verified(
        source_string: str, end_index: int, match_characters: str
    ) -> Tuple[int, int]:
        """
        Collect a sequence of the same character from a given starting point in
        a string going backwards towards the start of the string.

        Returns a tuple with the number of characters collected and the index of
        the first non-matching character and any extracted text.  If end_index is
        not within the source_string or the value of -1, this function will throw an AssertError.
        """
        index, extracted_string = (
            ParserHelper.collect_backwards_while_one_of_characters(
                source_string, end_index, match_characters
            )
        )
        assert index is not None
        assert extracted_string is not None
        return index, extracted_string

    @staticmethod
    def collect_until_character(
        source_string: str, start_index: int, match_character: str
    ) -> Tuple[Optional[int], Optional[str]]:
        """
        Collect a sequence of characters from a given starting point in a string until we hit a given character.

        Returns a tuple with the index of the first non-matching character and
        any extracted text
        OR
        a tuple of None values if the start_index is not within the source_string.
        """

        source_string_size = len(source_string)
        if not 0 <= start_index <= source_string_size:
            return None, None

        index = start_index
        while index < source_string_size and source_string[index] != match_character:
            index += 1

        return index, source_string[start_index:index]

    @staticmethod
    def collect_until_character_verified(
        source_string: str, start_index: int, match_character: str
    ) -> Tuple[int, str]:
        """
        Collect a sequence of characters from a given starting point in a string until we hit a given character.

        Returns a tuple with the index of the first non-matching character and
        any extracted text.  If start_index is not within the source_string,
        this function will throw an AssertError.
        """
        index, extracted_string = ParserHelper.collect_until_character(
            source_string, start_index, match_character
        )
        assert index is not None
        assert extracted_string is not None
        return index, extracted_string

    @staticmethod
    def collect_while_spaces(
        source_string: str, start_index: int
    ) -> Tuple[Optional[int], Optional[str]]:
        """
        Collect characters from a given starting point in a string as long
        as the character is either a string or a tab character.

        Returns a tuple with the index of the first non-matching character and
        any extracted text
        OR
        a tuple of None values if the start_index is not within the source_string.
        """
        return ParserHelper.collect_while_one_of_characters(
            source_string, start_index, ParserHelper.__normal_whitespace
        )

    @staticmethod
    def collect_while_spaces_verified(
        source_string: str, start_index: int
    ) -> Tuple[int, str]:
        """
        Collect characters from a given starting point in a string as long
        as the character is either a string or a tab character.

        Returns a tuple with the index of the first non-matching character and
        any extracted text.  If start_index is not within the source_string,
        this function will throw an AssertError.
        """
        index, extracted_string = ParserHelper.collect_while_spaces(
            source_string, start_index
        )
        assert index is not None
        assert extracted_string is not None
        return index, extracted_string

    @staticmethod
    def collect_while_one_of_characters(
        source_string: str, start_index: int, match_characters: str
    ) -> Tuple[Optional[int], Optional[str]]:
        """
        Collect a sequence of characters from a given starting point in a string as long
        as the character is one of the match characters.

        Returns a tuple with the index of the first non-matching character and any
        extracted text
        OR
        a tuple of None values if the start_index is not within the source_string.
        """

        source_string_size = len(source_string)
        if not 0 <= start_index <= source_string_size:
            return None, None

        index = start_index
        while index < source_string_size and source_string[index] in match_characters:
            index += 1

        return index, source_string[start_index:index]

    @staticmethod
    def collect_while_one_of_characters_verified(
        source_string: str, start_index: int, match_characters: str
    ) -> Tuple[int, str]:
        """
        Collect a sequence of characters from a given starting point in a string as long
        as the character is one of the match characters.

        Returns a tuple with the index of the first non-matching character and any
        extracted text.  If start_index is not within the source_string,
        this function will throw an AssertError.
        """
        index, extracted_string = ParserHelper.collect_while_one_of_characters(
            source_string, start_index, match_characters
        )
        assert index is not None
        assert extracted_string is not None
        return index, extracted_string

    @staticmethod
    def collect_until_one_of_characters(
        source_string: str, start_index: int, match_characters: str
    ) -> Tuple[Optional[int], Optional[str]]:
        """
        Collect a sequence of characters from a given starting point in a string until
        we hit one of a given set of characters.

        Returns a tuple with the index of the first non-matching character and any extracted text
        OR
        a tuple of None values if the start_index is not within the source_string.
        """

        source_string_size = len(source_string)
        if not 0 <= start_index <= source_string_size:
            return None, None

        index = start_index
        while (
            index < source_string_size and source_string[index] not in match_characters
        ):
            index += 1

        return index, source_string[start_index:index]

    @staticmethod
    def collect_until_one_of_characters_verified(
        source_string: str, start_index: int, match_characters: str
    ) -> Tuple[int, str]:
        """
        Collect a sequence of characters from a given starting point in a string until
        we hit one of a given set of characters.

        Returns a tuple with the index of the first non-matching character and any extracted text.
        If start_index is not within the source_string, this function will throw an AssertError.
        """

        index, extracted_string = ParserHelper.collect_until_one_of_characters(
            source_string, start_index, match_characters
        )
        assert index is not None
        assert extracted_string is not None
        return index, extracted_string

    @staticmethod
    def index_any_of(source_text: str, find_any: str, start_index: int = 0) -> int:
        """
        Determine if any of the specified characters are in the source string.
        """

        first_index = -1
        for next_character in find_any:
            found_index = source_text.find(next_character, start_index)
            if found_index != -1:
                first_index = (
                    found_index if first_index == -1 else min(first_index, found_index)
                )
                if first_index == 0:
                    break
        return first_index

    @staticmethod
    def replace_any_of(
        string_to_search_in: str, characters_to_search_for: str, replace_with: str
    ) -> str:
        """
        Replace any of a given set of characters with a given sequence.
        """

        start_index = 0
        index, ex_str = ParserHelper.collect_until_one_of_characters(
            string_to_search_in, start_index, characters_to_search_for
        )
        assert index is not None
        assert ex_str is not None
        string_to_search_in_size = len(string_to_search_in)
        replaced_parts: List[str] = []
        while index < string_to_search_in_size:
            replaced_parts.extend([ex_str, replace_with])
            start_index = index + 1
            index, ex_str = ParserHelper.collect_until_one_of_characters(
                string_to_search_in, start_index, characters_to_search_for
            )
            assert index is not None
            assert ex_str is not None
        replaced_parts.append(ex_str)
        return "".join(replaced_parts)

    @staticmethod
    def count_characters_in_text(text_to_examine: str, text_to_look_for: str) -> int:
        """
        Count the number of a given character in a given string.
        """
        original_length = len(text_to_examine)
        removed_length = len(text_to_examine.replace(text_to_look_for, ""))
        return original_length - removed_length

    @staticmethod
    def count_newlines_in_text(text_to_examine: str) -> int:
        """
        Count the number of new line characters in a given string.
        """
        return ParserHelper.count_characters_in_text(
            text_to_examine, ParserHelper.newline_character
        )

    @staticmethod
    def count_newlines_in_texts(*args: Any) -> int:
        """
        Count the number of new line characters in a given string.
        """
        return sum(
            ParserHelper.count_characters_in_text(
                next_argument, ParserHelper.newline_character
            )
            for next_argument in args
        )

    @staticmethod
    def calculate_deltas(text_to_analyze: str) -> Tuple[int, int]:
        """
        Calculate the deltas associated with a given string.
        """

        delta_line_number = 0
        if ParserHelper.newline_character in text_to_analyze:
            split_raw_tag = text_to_analyze.split(ParserHelper.newline_character)
            delta_line_number += len(split_raw_tag) - 1

            last_element = ParserHelper.__resolve_replacement_markers_from_text(
                split_raw_tag[-1]
            )
            last_element = ParserHelper.__remove_escapes_from_text(last_element)
            length_of_last_elements = len(last_element)

            delta_column_number = -(length_of_last_elements + 1)
        else:
            delta_column_number = len(text_to_analyze)
        return delta_line_number, delta_column_number

    # pylint: disable=too-many-arguments
    @staticmethod
    def recombine_string_with_whitespace(
        text_string: str,
        whitespace_string: str,
        start_index: int = 0,
        add_replace_marker_if_empty: bool = False,
        post_increment_index: bool = False,
        start_text_index: int = 1,
        add_whitespace_after: bool = False,
    ) -> Tuple[str, int]:
        """
        Properly recombine a text-string with a matching whitespace-string.
        """
        split_text_string, split_whitespace_string = (
            text_string.split(ParserHelper.newline_character),
            whitespace_string.split(ParserHelper.newline_character),
        )
        for i in range(start_text_index, len(split_text_string)):
            if not post_increment_index:
                start_index += 1
            ew_part = split_whitespace_string[start_index]
            if ew_part and add_replace_marker_if_empty:
                ew_part = ParserHelper.create_replace_with_nothing_marker(ew_part)
            split_text_string[i] = (
                f"{split_text_string[i]}{ew_part}"
                if add_whitespace_after
                else f"{ew_part}{split_text_string[i]}"
            )
            if post_increment_index:
                start_index += 1
        return ParserHelper.newline_character.join(split_text_string), start_index

    # pylint: enable=too-many-arguments

    @staticmethod
    def calculate_last_line(text_string: str) -> str:
        """
        Determine the last line of a multi-line string.
        """
        return text_string.split(ParserHelper.newline_character)[-1]

    @staticmethod
    def make_value_visible(value_to_modify: Any) -> str:
        """
        For the given value, turn it into a string if necessary, and then replace
        any known "invisible" characters with more visible strings.
        """
        return (
            str(value_to_modify)
            .replace(ParserHelper.__backspace_character, "\\b")
            .replace(ParserHelper.__alert_character, "\\a")
            .replace(ParserHelper.tab_character, "\\t")
            .replace(ParserHelper.newline_character, "\\n")
            .replace(ParserHelper.whitespace_split_character, "\\x02")
            .replace(ParserHelper.replace_noop_character, "\\x03")
            .replace(ParserHelper.escape_character, "\\x05")
            .replace("\\x07", "\\a")
            .replace("\\x08", "\\b")
        )

    @staticmethod
    def make_whitespace_visible(value_to_modify: str) -> str:
        """
        For the given value, turn it into a string if necessary, and then replace
        any known whitespace characters with more visible strings.
        """
        # sourcery skip: remove-unnecessary-cast
        return (
            str(value_to_modify)
            .replace(ParserHelper.newline_character, "\\n")
            .replace(" ", "\\s")
        )

    @staticmethod
    def valid_characters_to_escape() -> str:
        """
        List of valid characters that can be escaped.
        """
        return ParserHelper.all_escape_characters

    @staticmethod
    def escape_special_characters(string_to_escape: str) -> str:
        """
        Build another string that has any special characters in the argument escaped.
        """
        string_parts = []
        for next_char_index, next_character in enumerate(string_to_escape):
            if ParserHelper.is_character_at_index_one_of(
                string_to_escape, next_char_index, ParserHelper.all_escape_characters
            ):
                string_parts.append(ParserHelper.escape_character)
            string_parts.append(next_character)
        return "".join(string_parts)

    @staticmethod
    def __remove_backspaces_from_text(token_text: str) -> str:
        """
        Remove any backspaces from the text.
        """
        adjusted_text_token = token_text[:]
        next_backspace_index = ParserHelper.__find_with_escape(
            adjusted_text_token, ParserHelper.__backspace_character, 0
        )
        while next_backspace_index != -1:
            adjusted_text_token = (
                adjusted_text_token[:next_backspace_index]
                + adjusted_text_token[next_backspace_index + 1 :]
            )
            next_backspace_index = ParserHelper.__find_with_escape(
                adjusted_text_token,
                ParserHelper.__backspace_character,
                next_backspace_index,
            )
        return adjusted_text_token

    @staticmethod
    def resolve_backspaces_from_text(token_text: str) -> str:
        """
        Deal with any backslash encoding in text with backspaces.
        """
        adjusted_text_token = token_text[:]
        next_backspace_index = ParserHelper.__find_with_escape(
            adjusted_text_token, ParserHelper.__backspace_character, 0
        )
        while next_backspace_index != -1:
            adjusted_text_token = (
                adjusted_text_token[: next_backspace_index - 1]
                + adjusted_text_token[next_backspace_index + 1 :]
            )
            next_backspace_index = ParserHelper.__find_with_escape(
                adjusted_text_token,
                ParserHelper.__backspace_character,
                next_backspace_index,
            )
        return adjusted_text_token

    @staticmethod
    def create_replacement_markers(
        replace_this_string: str, with_this_string: str
    ) -> str:
        """
        Create a replacement marker indicating that the first string is being replaced
        by the second string.
        """
        return (
            f"{ParserHelper.__alert_character}{replace_this_string}{ParserHelper.__alert_character}"
            + f"{with_this_string}{ParserHelper.__alert_character}"
        )

    @staticmethod
    def create_replace_with_nothing_marker(replace_this_string: str) -> str:
        """
        Create a replacement marker of the given string with the noop character.
        """
        return (
            f"{ParserHelper.__alert_character}{replace_this_string}{ParserHelper.__alert_character}"
            + f"{ParserHelper.replace_noop_character}{ParserHelper.__alert_character}"
        )

    @staticmethod
    def __remove_sequence_from_text(token_text: str, sequence_to_remove: str) -> str:
        """
        Resolve the specific character out of the text string.
        """
        adjusted_text_token = token_text[:]
        next_backspace_index = ParserHelper.__find_with_escape(
            adjusted_text_token, sequence_to_remove, 0
        )
        while next_backspace_index != -1:
            adjusted_text_token = (
                adjusted_text_token[:next_backspace_index]
                + adjusted_text_token[next_backspace_index + 1 :]
            )
            next_backspace_index = ParserHelper.__find_with_escape(
                adjusted_text_token, sequence_to_remove, next_backspace_index
            )
        return adjusted_text_token

    @staticmethod
    def resolve_noops_from_text(token_text: str) -> str:
        """
        Resolve the replacement noop character out of the text string.
        """
        return ParserHelper.__remove_sequence_from_text(
            token_text, ParserHelper.replace_noop_character
        )

    @staticmethod
    def __resolve_escapes_from_text(token_text: str) -> str:
        """
        Resolve any escapes from the text, leaving only what they escaped.
        """
        adjusted_text_token = token_text[:]
        next_backspace_index = ParserHelper.__find_with_escape(
            adjusted_text_token, ParserHelper.escape_character, 0
        )
        while next_backspace_index != -1:
            adjusted_text_token = (
                adjusted_text_token[:next_backspace_index]
                + adjusted_text_token[next_backspace_index + 1 :]
            )
            next_backspace_index = ParserHelper.__find_with_escape(
                adjusted_text_token,
                ParserHelper.escape_character,
                next_backspace_index + 1,
            )
        return adjusted_text_token

    @staticmethod
    def __remove_escapes_from_text(token_text: str) -> str:
        """
        Remove any escape characters from the text.
        """
        return ParserHelper.__resolve_escapes_from_text(token_text)

    @staticmethod
    def is_character_replacement_marker(
        source_string: str, index_in_string: int
    ) -> bool:
        """
        Determine if the specified character is a replacement marker character.
        """
        return (
            0 <= index_in_string < len(source_string)
            and source_string[index_in_string] == ParserHelper.__alert_character
        )

    @staticmethod
    def get_replacement_indices(
        source_string: str, index_in_string: int
    ) -> Tuple[int, int, int]:
        """
        Find the next replacement character index, as well as the middle and end indices
        for the replacement.
        """
        start_replacement_index = ParserHelper.__find_with_escape(
            source_string, ParserHelper.__alert_character, index_in_string
        )
        middle_replacement_index = -1
        end_replacement_index = -1
        if start_replacement_index != -1:
            middle_replacement_index = source_string.find(
                ParserHelper.__alert_character, start_replacement_index + 1
            )
            if middle_replacement_index != -1:
                end_replacement_index = source_string.find(
                    ParserHelper.__alert_character, middle_replacement_index + 1
                )
        return start_replacement_index, middle_replacement_index, end_replacement_index

    @staticmethod
    def __resolve_replacement_markers_from_text(main_text: str) -> str:
        """
        Resolve the alert characters (i.e. replacement markers) out of the text string.
        """
        start_replacement_index = ParserHelper.__find_with_escape(
            main_text, ParserHelper.__alert_character, 0
        )
        while start_replacement_index != -1:
            middle_replacement_index = main_text.index(
                ParserHelper.__alert_character, start_replacement_index + 1
            )
            end_replacement_index = main_text.index(
                ParserHelper.__alert_character, middle_replacement_index + 1
            )

            replace_text = main_text[
                start_replacement_index + 1 : middle_replacement_index
            ]

            # It is possible to have one level of nesting, so deal with it.
            if middle_replacement_index + 1 == end_replacement_index:
                inner_start_replacement_index = main_text.index(
                    ParserHelper.__alert_character, end_replacement_index + 1
                )
                inner_middle_replacement_index = main_text.index(
                    ParserHelper.__alert_character, inner_start_replacement_index + 1
                )
                inner_end_replacement_index = main_text.index(
                    ParserHelper.__alert_character, inner_middle_replacement_index + 1
                )
                assert inner_middle_replacement_index + 1 == inner_end_replacement_index
                end_replacement_index = inner_end_replacement_index

            length_before_mod = len(main_text)
            main_text = (
                (
                    main_text[:start_replacement_index]
                    + replace_text
                    + main_text[end_replacement_index + 1 :]
                )
                if start_replacement_index
                else (replace_text + main_text[end_replacement_index + 1 :])
            )
            length_after_mod = len(main_text)
            start_index = (
                end_replacement_index + 1 + (length_after_mod - length_before_mod)
            )
            start_replacement_index = ParserHelper.__find_with_escape(
                main_text, ParserHelper.__alert_character, start_index
            )
        return main_text

    @staticmethod
    def __find_with_escape(
        adjusted_text_token: str, find_char: str, start_index: int
    ) -> int:
        found_index = -1
        while start_index < len(adjusted_text_token):
            start_replacement_index = adjusted_text_token.find(find_char, start_index)
            if (
                start_replacement_index != -1
                and start_replacement_index > 0
                and adjusted_text_token[start_replacement_index - 1]
                == ParserHelper.escape_character
            ):
                start_index = start_replacement_index + 1
            else:
                found_index = start_replacement_index
                break
        return found_index

    @staticmethod
    def __resolve_references_from_text(adjusted_text_token: str) -> str:
        """
        The alert characters signal that a replacement has occurred, so make sure
        we take the right text from the replacement.
        """
        start_replacement_index = ParserHelper.__find_with_escape(
            adjusted_text_token, ParserHelper.__alert_character, 0
        )
        while start_replacement_index != -1:
            middle_replacement_index = adjusted_text_token.index(
                ParserHelper.__alert_character, start_replacement_index + 1
            )
            end_replacement_index = adjusted_text_token.index(
                ParserHelper.__alert_character, middle_replacement_index + 1
            )

            if middle_replacement_index + 1 == end_replacement_index:
                inner_start_replacement_index = adjusted_text_token.index(
                    ParserHelper.__alert_character, end_replacement_index + 1
                )
                inner_middle_replacement_index = adjusted_text_token.index(
                    ParserHelper.__alert_character, inner_start_replacement_index + 1
                )
                inner_end_replacement_index = adjusted_text_token.index(
                    ParserHelper.__alert_character, inner_middle_replacement_index + 1
                )
                replace_text = adjusted_text_token[
                    inner_start_replacement_index + 1 : inner_middle_replacement_index
                ]
                assert inner_middle_replacement_index + 1 == inner_end_replacement_index
                end_replacement_index = inner_end_replacement_index
            else:
                replace_text = adjusted_text_token[
                    middle_replacement_index + 1 : end_replacement_index
                ]

            length_before_mod = len(adjusted_text_token)
            adjusted_text_token = (
                (
                    adjusted_text_token[:start_replacement_index]
                    + replace_text
                    + adjusted_text_token[end_replacement_index + 1 :]
                )
                if start_replacement_index
                else replace_text + adjusted_text_token[end_replacement_index + 1 :]
            )
            length_after_mod = len(adjusted_text_token)
            start_index = (
                end_replacement_index + 1 + (length_after_mod - length_before_mod)
            )
            start_replacement_index = ParserHelper.__find_with_escape(
                adjusted_text_token, ParserHelper.__alert_character, start_index
            )
        return adjusted_text_token

    @staticmethod
    def resolve_all_from_text(text_to_resolve: str) -> str:
        """
        Combination to resolve all of these special characters from the text.
        """
        resolved_text = ParserHelper.resolve_backspaces_from_text(text_to_resolve)
        resolved_text = ParserHelper.__resolve_references_from_text(resolved_text)
        resolved_text = ParserHelper.resolve_noops_from_text(resolved_text)
        return ParserHelper.__resolve_escapes_from_text(resolved_text)

    @staticmethod
    def remove_all_from_text(text_to_remove: str, include_noops: bool = False) -> str:
        """
        Combination to remove all of these special characters from the text.
        """
        removed_text = ParserHelper.__remove_backspaces_from_text(text_to_remove)
        removed_text = ParserHelper.__resolve_replacement_markers_from_text(
            removed_text
        )
        if include_noops:
            removed_text = ParserHelper.resolve_noops_from_text(removed_text)
        return ParserHelper.__remove_escapes_from_text(removed_text)

    @staticmethod
    def repeat_string(string_to_repeat: str, repeat_count: int) -> str:
        """
        Repeat the given character the specified number of times.
        """
        return "".rjust(repeat_count, string_to_repeat)

    @staticmethod
    def find_nth_occurrence(search_in: str, search_for: str, nth: int) -> int:
        """
        Search for the nth (1-based) occurrence of the search_for
        string within the search_in string.
        """

        did_find_last, found_index, start_index = True, -1, 0
        while nth > 0 and did_find_last:
            found_index = search_in.find(search_for, start_index)
            did_find_last = found_index != -1
            if did_find_last:
                start_index = found_index + 1
                nth -= 1
        return found_index if did_find_last else -1

    @staticmethod
    def adjust_for_newlines(
        source_string: str, start_index: int, end_index: int
    ) -> Tuple[int, int]:
        """
        Given the various parameters, determine the column and
        line offsets implied by newline characters in the
        specified text.
        """

        col_adjust, line_adjust, newline_index = (
            end_index,
            0,
            source_string.find(ParserHelper.newline_character, start_index),
        )
        while newline_index != -1 and newline_index < end_index:
            line_adjust += 1
            col_adjust, newline_index = -(
                end_index - newline_index
            ), source_string.find(ParserHelper.newline_character, newline_index + 1)
        return col_adjust, line_adjust


# pylint: enable=too-many-public-methods
